from snovault import (
    calculated_property,
    collection,
    load_schema,
)
from .file import File, file_workflow_run_embeds


@collection(
    name='files-processed',
    properties={
        'title': 'Processed Files',
        'description': 'Listing of Processed Files',
    })
class FileProcessed(File):
    """Collection for individual processed files."""
    item_type = 'file_processed'
    schema = load_schema('encoded_core:schemas/file_processed.json')
    embedded_list = File.embedded_list + file_workflow_run_embeds
    rev = dict(File.rev, **{
        'workflow_run_inputs': ('WorkflowRun', 'input_files.value'),
        'workflow_run_outputs': ('WorkflowRun', 'output_files.value'),
    })
    aggregated_items = {
        "last_modified": [
            "date_modified"
        ],
    }

    @classmethod
    def get_bucket(cls, registry):
        return registry.settings['file_wfout_bucket']

    @calculated_property(schema={
        "title": "Input of Workflow Runs",
        "description": "All workflow runs that this file serves as an input to",
        "type": "array",
        "items": {
            "title": "Input of Workflow Run",
            "type": ["string", "object"],
            "linkTo": "WorkflowRun"
        }
    })
    def workflow_run_inputs(self, request, disable_wfr_inputs=False):
        # switch this calc prop off for some processed files, i.e. control exp files
        if not disable_wfr_inputs:
            return self.rev_link_atids(request, "workflow_run_inputs")
        else:
            return []

    @calculated_property(schema={
        "title": "Output of Workflow Runs",
        "description": "All workflow runs that this file serves as an output from",
        "type": "array",
        "items": {
            "title": "Output of Workflow Run",
            "type": "string",
            "linkTo": "WorkflowRun"
        }
    })
    def workflow_run_outputs(self, request):
        return self.rev_link_atids(request, "workflow_run_outputs")

    # processed files don't want md5 as unique key
    def unique_keys(self, properties):
        keys = super(FileProcessed, self).unique_keys(properties)
        if keys.get('alias'):
            keys['alias'] = [k for k in keys['alias'] if not k.startswith('md5:')]
        return keys



import datetime
import pytz
import structlog

from pyramid.httpexceptions import (
    HTTPForbidden,
    HTTPTemporaryRedirect,
    HTTPNotFound,
)
from pyramid.response import Response
from pyramid.settings import asbool
from pyramid.view import view_config
from snovault import (
    AfterModified,
    BeforeModified,
)
from snovault.elasticsearch import ELASTIC_SEARCH
from snovault.schema_utils import schema_validator
from snovault.util import debug_log
from snovault.validators import (
    validate_item_content_post,
    validate_item_content_put,
    validate_item_content_patch,
    validate_item_content_in_place,
    no_validate_item_content_post,
    no_validate_item_content_put,
    no_validate_item_content_patch
)
from urllib.parse import (
    parse_qs,
    urlparse,
)
from snovault.authentication import session_properties
from snovault.search.search import make_search_subreq
from snovault.util import check_user_is_logged_in, make_s3_client
from snovault.types.base import (
    get_item_or_none,
    collection_add,
    item_edit,
)
from .types.file import File, external_creds


log = structlog.getLogger(__name__)


def includeme(config):
    config.scan(__name__)


@view_config(name='upload', context=File, request_method='GET',
             permission='edit')
@debug_log
def get_upload(context, request):
    external = context.propsheets.get('external', {})
    upload_credentials = external.get('upload_credentials')
    # Show s3 location info for files originally submitted to EDW.

    if upload_credentials is None and external.get('service') == 's3':
        upload_credentials = {
            'upload_url': 's3://{bucket}/{key}'.format(**external),
        }
    return {
        '@graph': [{
            '@id': request.resource_path(context),
            'upload_credentials': upload_credentials,
            'extra_files_creds': context.extra_files_creds(),
        }],
    }


@view_config(name='upload', context=File, request_method='POST',
             permission='edit', validators=[schema_validator({"type": "object"})])
@debug_log
def post_upload(context, request):
    properties = context.upgrade_properties()
    if properties['status'] not in ('uploading', 'to be uploaded by workflow', 'upload failed'):
        raise HTTPForbidden('status must be "uploading" to issue new credentials')
    # accession_or_external = properties.get('accession')
    external = context.propsheets.get('external', None)

    if external is None:
        # Handle objects initially posted as another state.
        bucket = request.registry.settings['file_upload_bucket']
        # maybe this should be properties.uuid
        uuid = context.uuid
        file_format = get_item_or_none(request, properties.get('file_format'), 'file-formats')
        try:
            file_extension = '.' + file_format.get('standard_file_extension')
        except AttributeError:
            file_extension = ''

        key = '{uuid}/{accession}.{file_extension}'.format(
            file_extension=file_extension, uuid=uuid, **properties)

    elif external.get('service') == 's3':
        bucket = external['bucket']
        key = external['key']
    else:
        raise ValueError(external.get('service'))

    # remove the path from the file name and only take first 32 chars
    name = None
    if properties.get('filename'):
        name = properties.get('filename').split('/')[-1][:32]
    profile_name = request.registry.settings.get('file_upload_profile_name')
    creds = external_creds(bucket, key, name, profile_name)
    # in case we haven't uploaded a file before
    context.propsheets['external'] = creds

    new_properties = properties.copy()
    if properties['status'] == 'upload failed':
        new_properties['status'] = 'uploading'

    registry = request.registry
    registry.notify(BeforeModified(context, request))
    context.update(new_properties, {'external': creds})
    registry.notify(AfterModified(context, request))

    rendered = request.embed('/%s/@@object' % context.uuid, as_user=True)
    result = {
        'status': 'success',
        '@type': ['result'],
        '@graph': [rendered],
    }
    return result


def is_file_to_download(properties, file_format, expected_filename=None):
    try:
        file_extension = '.' + file_format.get('standard_file_extension')
    except AttributeError:
        file_extension = ''
    accession_or_external = properties.get('accession') or properties.get('external_accession')
    if not accession_or_external:
        return False
    filename = '{accession}{file_extension}'.format(
        accession=accession_or_external, file_extension=file_extension)
    if expected_filename is None:
        return filename
    elif expected_filename != filename:
        return False
    else:
        return filename


@view_config(name='download', context=File, request_method='GET',
             permission='view', subpath_segments=[0, 1])
def download(context, request):
    """ File download route. Generates a pre-signed S3 URL for the object that expires eventually. """
    check_user_is_logged_in(request)

    # first check for restricted status
    try:
        user_props = session_properties(context, request)
    except Exception as e:
        user_props = {'error': str(e)}
    tracking_values = {'user_agent': request.user_agent, 'remote_ip': request.remote_addr,
                       'user_uuid': user_props.get('details', {}).get('uuid', 'anonymous'),
                       'request_path': request.path_info, 'request_headers': str(dict(request.headers))}

    # proxy triggers if we should use Axel-redirect, useful for s3 range byte queries
    try:
        use_download_proxy = request.client_addr not in request.registry['aws_ipset']
    except TypeError:
        # this fails in testing due to testapp not having ip
        use_download_proxy = False

    # with extra_files the user may be trying to download the main file
    # or one of the files in extra files, the following logic will
    # search to find the "right" file and redirect to a download link for that one
    properties = context.upgrade_properties()
    file_format = get_item_or_none(request, properties.get('file_format'), 'file-formats')
    _filename = None
    if request.subpath:
        _filename, = request.subpath
    filename = is_file_to_download(properties, file_format, _filename)
    if not filename:
        found = False
        for extra in properties.get('extra_files', []):
            eformat = get_item_or_none(request, extra.get('file_format'), 'file-formats')
            filename = is_file_to_download(extra, eformat, _filename)
            if filename:
                found = True
                properties = extra
                external = context.propsheets.get('external' + eformat.get('uuid'))
                if eformat is not None:
                    tracking_values['file_format'] = eformat.get('file_format')
                break
        if not found:
            raise HTTPNotFound(_filename)
    else:
        external = context.propsheets.get('external', {})
        if file_format is not None:
            tracking_values['file_format'] = file_format.get('file_format')
    tracking_values['filename'] = filename

    if not external:
        external = context.build_external_creds(request.registry, context.uuid, properties)
    if external.get('service') == 's3':
        external_bucket = external['bucket']
        wfout_bucket = request.registry.settings['file_wfout_bucket']
        files_bucket = request.registry.settings['file_upload_bucket']
        if external_bucket not in [wfout_bucket, files_bucket]:
            if 'wfout' not in external_bucket:
                external_bucket = files_bucket
            else:
                external_bucket = wfout_bucket
            log.error(f'Encountered s3 bucket mismatch - ignoring metadata value {external_bucket}'
                      f' and using registry value {external_bucket}')
        conn = make_s3_client()
        param_get_object = {
            'Bucket': external_bucket,
            'Key': external['key'],
            'ResponseContentDisposition': 'attachment; filename=' + filename
        }
        if 'Range' in request.headers:
            tracking_values['range_query'] = True
            param_get_object.update({'Range': request.headers.get('Range')})
        else:
            tracking_values['range_query'] = False
        location = conn.generate_presigned_url(
            ClientMethod='get_object',
            Params=param_get_object,
            ExpiresIn=36*60*60
        )
    else:
        raise ValueError(external.get('service'))

    # tracking_values['experiment_type'] = get_file_experiment_type(request, context, properties)
    # # create a tracking_item to track this download
    # tracking_item = {'status': 'in review by lab', 'tracking_type': 'download_tracking',
    #                  'download_tracking': tracking_values}
    # try:
    #     TrackingItem.create_and_commit(request, tracking_item, clean_headers=True)
    # except Exception as e:
    #     log.error('Cannot create TrackingItem on download of %s' % context.uuid, error=str(e))

    if asbool(request.params.get('soft')):
        expires = int(parse_qs(urlparse(location).query)['Expires'][0])
        return {
            '@type': ['SoftRedirect'],
            'location': location,
            'expires': datetime.datetime.fromtimestamp(expires, pytz.utc).isoformat(),
        }

    if 'Range' in request.headers:
        try:
            response_body = conn.get_object(**param_get_object)
        except Exception as e:
            raise e
        response_dict = {
            'body': response_body.get('Body').read(),
            # status_code : 206 if partial, 200 if the range covers whole file
            'status_code': response_body.get('ResponseMetadata').get('HTTPStatusCode'),
            'accept_ranges': response_body.get('AcceptRanges'),
            'content_length': response_body.get('ContentLength'),
            'content_range': response_body.get('ContentRange')
        }
        return Response(**response_dict)

    # We don't use X-Accel-Redirect here so that client behaviour is similar for
    # both aws and non-aws users.
    if use_download_proxy:
        location = request.registry.settings.get('download_proxy', '') + str(location)

    # 307 redirect specifies to keep original method
    raise HTTPTemporaryRedirect(location=location)


def validate_file_format_validity_for_file_type(context, request):
    """Check if the specified file format (e.g. fastq) is allowed for the file type (e.g. FileFastq).
    """
    data = request.json
    if 'file_format' in data:
        file_format_item = get_item_or_none(request, data['file_format'], 'file-formats')
        if not file_format_item:
            # item level validation will take care of generating the error
            return
        file_format_name = file_format_item['file_format']
        allowed_types = file_format_item.get('valid_item_types', [])
        file_type = context.type_info.name
        if file_type not in allowed_types:
            msg = 'File format {} is not allowed for {}'.format(file_format_name, file_type)
            request.errors.add('body', 'File: invalid format', msg)
        else:
            request.validated.update({})


def validate_file_filename(context, request):
    """ validator for filename field """
    found_match = False
    data = request.json
    if 'filename' not in data:
        # see if there is an existing file_name
        filename = context.properties.get('filename')
        if not filename:
            return
    else:
        filename = data['filename']
    ff = data.get('file_format')
    if not ff:
        ff = context.properties.get('file_format')
    file_format_item = get_item_or_none(request, ff, 'file-formats')
    if not file_format_item:
        msg = 'Problem getting file_format for %s' % filename
        request.errors.add('body', 'File: no format', msg)
        return
    msg = None
    try:
        file_extensions = [file_format_item.get('standard_file_extension')]
        if file_format_item.get('other_allowed_extensions'):
            file_extensions.extend(file_format_item.get('other_allowed_extensions'))
            file_extensions = list(set(file_extensions))
    except (AttributeError, TypeError):
        msg = 'Problem getting file_format for %s' % filename
    else:
        if file_format_item.get('file_format') == 'other':
            found_match = True
        elif not file_extensions:  # this shouldn't happen
            pass
        for extension in file_extensions:
            if filename[-(len(extension) + 1):] == '.' + extension:
                found_match = True
                break
    if found_match:
        request.validated.update({})
    else:
        if not msg:
            msg = ["'." + ext + "'" for ext in file_extensions]
            msg = ', '.join(msg)
            msg = ('Filename %s extension does not agree with specified file format. '
                   'Valid extension(s): %s' % (filename, msg))
        request.errors.add('body', 'File: invalid extension', msg)


def validate_processed_file_unique_md5_with_bypass(context, request):
    """validator to check md5 on processed files, unless you tell it not to"""
    # skip validator if not file processed
    if context.type_info.item_type != 'file_processed':
        return
    data = request.json
    if 'md5sum' not in data or not data['md5sum']:
        return
    if 'force_md5' in request.query_string:
        return
    # we can of course patch / put to ourselves the same md5 we previously had
    if context.properties.get('md5sum') == data['md5sum']:
        return

    if ELASTIC_SEARCH in request.registry:
        search = make_search_subreq(request, '/search/?type=File&md5sum=%s' % data['md5sum'])
        search_resp = request.invoke_subrequest(search, True)
        if search_resp.status_int < 400:
            # already got this md5
            found = search_resp.json['@graph'][0]['accession']
            request.errors.add('body', 'File: non-unique md5sum', 'md5sum %s already exists for accession %s' %
                               (data['md5sum'], found))
    else:  # find it in the database
        conn = request.registry['connection']
        res = conn.get_by_json('md5sum', data['md5sum'], 'file_processed')
        if res is not None:
            # md5 already exists
            found = res.properties['accession']
            request.errors.add('body', 'File: non-unique md5sum', 'md5sum %s already exists for accession %s' %
                               (data['md5sum'], found))


def validate_processed_file_produced_from_field(context, request):
    """validator to make sure that the values in the
    produced_from field are valid file identifiers"""
    # skip validator if not file processed
    if context.type_info.item_type != 'file_processed':
        return
    data = request.json
    if 'produced_from' not in data:
        return
    files_ok = True
    files2chk = data['produced_from']
    for i, f in enumerate(files2chk):
        try:
            fid = get_item_or_none(request, f, 'files').get('uuid')
        except AttributeError:
            files_ok = False
            request.errors.add('body', 'File: invalid produced_from id', "'%s' not found" % f)
            # bad_files.append(f)
        else:
            if not fid:
                files_ok = False
                request.errors.add('body', 'File: invalid produced_from id', "'%s' not found" % f)

    if files_ok:
        request.validated.update({})


def validate_extra_file_format(context, request):
    """validator to check to be sure that file_format of extrafile is not the
       same as the file and is a known format for the schema
    """
    files_ok = True
    data = request.json
    if not data.get('extra_files'):
        return
    extras = data['extra_files']
    # post should always have file_format as it is required patch may or may not
    ff = data.get('file_format')
    if not ff:
        ff = context.properties.get('file_format')
    file_format_item = get_item_or_none(request, ff, 'file-formats')
    if not file_format_item or 'standard_file_extension' not in file_format_item:
        request.errors.add('body', 'File: no extra_file format', "Can't find parent file format for extra_files")
        return
    parent_format = file_format_item['uuid']
    schema_eformats = file_format_item.get('extrafile_formats')
    if not schema_eformats:  # means this parent file shouldn't have any extra files
        request.errors.add(
            'body', 'File: invalid extra files',
            "File with format %s should not have extra_files" % file_format_item.get('file_format')
        )
        return
    else:
        valid_ext_formats = []
        for ok_format in schema_eformats:
            ok_format_item = get_item_or_none(request, ok_format, 'file-formats')
            try:
                off_uuid = ok_format_item.get('uuid')
            except AttributeError:
                raise Exception("FileFormat Item %s contains unknown FileFormats"
                                " in the extrafile_formats property" % file_format_item.get('uuid'))
            valid_ext_formats.append(off_uuid)
    seen_ext_formats = []
    # formats = request.registry['collections']['FileFormat']
    for i, ef in enumerate(extras):
        eformat = ef.get('file_format')
        if eformat is None:
            return  # will fail the required extra_file.file_format
        eformat_item = get_item_or_none(request, eformat, 'file-formats')
        try:
            ef_uuid = eformat_item.get('uuid')
        except AttributeError:
            request.errors.add(
                'body', 'File: invalid extra_file format', "'%s' not a valid or known file format" % eformat
            )
            files_ok = False
            break
        if ef_uuid in seen_ext_formats:
            request.errors.add(
                'body', 'File: invalid extra_file formats',
                "Multple extra files with '%s' format cannot be submitted at the same time" % eformat
            )
            files_ok = False
            break
        else:
            seen_ext_formats.append(ef_uuid)
        if ef_uuid == parent_format:
            request.errors.add(
                'body', 'File: invalid extra_file formats',
                "'%s' format cannot be the same for file and extra_file" % file_format_item.get('file_format')
            )
            files_ok = False
            break

        if ef_uuid not in valid_ext_formats:
            request.errors.add(
                'body', 'File: invalid extra_file formats',
                "'%s' not a valid extrafile_format for '%s'" % (eformat, file_format_item.get('file_format'))
            )
            files_ok = False
    if files_ok:
        request.validated.update({})


def build_drs_object_from_props(drs_object_base, props):
    """ Takes in base properties for a DRS object we expect on all items and expands them if corresponding
        properties exist
    """
    # fields that match exactly in name and structure
    for exact_field in [
        'description',
    ]:
        if exact_field in props:
            drs_object_base[exact_field] = props[exact_field]

    # size is required by DRS so take it or default to 0
    drs_object_base['size'] = props.get('file_size', 0)

    # use system uuid as alias
    drs_object_base['aliases'] = [props['uuid']]

    # fields that are mapped to different names/structure
    if 'content_md5sum' in props:
        drs_object_base['checksums'] = [
            {
                'checksum': props['content_md5sum'],
                'type': 'md5'
            }
        ]
        # use md5sum as version
        drs_object_base['version'] = props['content_md5sum']
    if 'filename' in props:
        drs_object_base['name'] = props['filename']
    if 'last_modified' in props:
        drs_object_base['updated_time'] = props['last_modified']['date_modified']
    return drs_object_base


@view_config(name='drs', context=File, request_method='GET',
             permission='view', subpath_segments=[0, 1])
def drs(context, request):
    """ DRS object implementation for file. """
    rendered_object = request.embed(str(context.uuid), '@@object', as_user=True)
    accession = rendered_object['accession']
    drs_object_base = {
        'id': rendered_object['@id'],
        'created_time': rendered_object['date_created'],
        'drs_id': accession,
        'self_uri': f'drs://{request.host}{request.path}',
        'access_methods': [
            {
                # always prefer https
                'access_url': {
                    'url': f'https://{request.host}/{accession}/@@download'
                },
                'type': 'https'
            },
            {
                # but provide http as well in case we are not on prod
                'access_url': {
                    'url': f'http://{request.host}/{accession}/@@download'
                },
                'type': 'http'
            },
        ]
    }
    return build_drs_object_from_props(drs_object_base, rendered_object)


@view_config(context=File.Collection, permission='add', request_method='POST',
             validators=[validate_item_content_post,
                         validate_file_filename,
                         validate_extra_file_format,
                         validate_file_format_validity_for_file_type,
                         validate_processed_file_unique_md5_with_bypass,
                         validate_processed_file_produced_from_field])
@view_config(context=File.Collection, permission='add_unvalidated', request_method='POST',
             validators=[no_validate_item_content_post],
             request_param=['validate=false'])
@debug_log
def file_add(context, request, render=None):
    return collection_add(context, request, render)


@view_config(context=File, permission='edit', request_method='PUT',
             validators=[validate_item_content_put,
                         validate_file_filename,
                         validate_extra_file_format,
                         validate_file_format_validity_for_file_type,
                         validate_processed_file_unique_md5_with_bypass,
                         validate_processed_file_produced_from_field])
@view_config(context=File, permission='edit', request_method='PATCH',
             validators=[validate_item_content_patch,
                         validate_file_filename,
                         validate_extra_file_format,
                         validate_file_format_validity_for_file_type,
                         validate_processed_file_unique_md5_with_bypass,
                         validate_processed_file_produced_from_field])
@view_config(context=File, permission='edit_unvalidated', request_method='PUT',
             validators=[no_validate_item_content_put],
             request_param=['validate=false'])
@view_config(context=File, permission='edit_unvalidated', request_method='PATCH',
             validators=[no_validate_item_content_patch],
             request_param=['validate=false'])
@view_config(context=File, permission='index', request_method='GET',
             validators=[validate_item_content_in_place,
                         validate_file_filename,
                         validate_extra_file_format,
                         validate_file_format_validity_for_file_type,
                         validate_processed_file_unique_md5_with_bypass,
                         validate_processed_file_produced_from_field],
             request_param=['check_only=true'])
@debug_log
def file_edit(context, request, render=None):
    return item_edit(context, request, render)



import os
from pyramid.config import Configurator
from pyramid.settings import asbool

from snovault.app import session, configure_dbsession
from snovault.elasticsearch import APP_FACTORY
from snovault.elasticsearch.interfaces import INVALIDATION_SCOPE_ENABLED
from .local_roles import LocalRolesAuthorizationPolicy


APP_VERSION_REGISTRY_KEY = 'encoded-core.app_version'


def includeme(config):
    """include me method."""
    config.scan()


def app_version(config):
    if not config.registry.settings.get(APP_VERSION_REGISTRY_KEY):
        # we update version as part of deployment process `deploy_beanstalk.py`
        # but if we didn't check env then git
        version = os.environ.get("ENCODED_VERSION", "test")
        config.registry.settings[APP_VERSION_REGISTRY_KEY] = version


def main(global_config, **local_config):
    """ Returns a pyramid wsgi application that can be used as encoded-core - should only
        be used in testing
    """
    settings = global_config
    settings.update(local_config)

    # enable invalidation scope
    settings[INVALIDATION_SCOPE_ENABLED] = True

    config = Configurator(settings=settings)

    config.registry[APP_FACTORY] = main  # used by mp_indexer
    config.include(app_version)

    config.include('pyramid_multiauth')  # must be before calling set_authorization_policy
    # Override default authz policy set by pyramid_multiauth
    config.set_authorization_policy(LocalRolesAuthorizationPolicy())
    config.include(session)

    # must include, as tm.attempts was removed from pyramid_tm
    config.include('pyramid_retry')

    # for CGAP, always enable type=nested mapping
    # NOTE: this MUST occur prior to including Snovault, otherwise it will not work
    config.add_settings({'mappings.use_nested': True})
    config.include(configure_dbsession)
    config.include('snovault')
    config.commit()  # commit so search can override listing

    if 'elasticsearch.server' in config.registry.settings:
        config.include('snovault.elasticsearch')
        config.include('.search.search')
        config.include('.search.compound_search')  # could make enabling configurable

    if asbool(settings.get('testing', False)):
        config.include('snovault.tests.testing_views')
        config.include('snovault.root')

    config.include('.upgrade')

    app = config.make_wsgi_app()
    return app



from dcicutils.misc_utils import filtered_warnings
from pyramid.httpexceptions import (  # 301-307 redirect code response
    HTTPMovedPermanently,
    HTTPFound,
    HTTPSeeOther,
    HTTPTemporaryRedirect
)
from pyramid.view import view_config
from snovault import COLLECTIONS, CONNECTION
from snovault.resource_views import item_view_page
from snovault.util import debug_log
from snovault.validators import (
    validate_item_content_post,
    validate_item_content_put,
    validate_item_content_patch,
    validate_item_content_in_place,
    no_validate_item_content_post,
    no_validate_item_content_put,
    no_validate_item_content_patch
)
from urllib.parse import urlparse, urlencode
from snovault.search.search import get_iterable_search_results
from snovault.types.base import collection_add, item_edit
from .types.page import Page


def includeme(config):
    with filtered_warnings("ignore", category=DeprecationWarning):
        config.add_route(
            'staticpage',
            '/*subpath',
            # TODO: Replace custom_predicates=[is_static_page] with something more modern.
            # The custom_predicates needs to be rewritten.
            # Although there is a complex rewrite using .add_route_predicate,
            # the simpler case of just using .add_static_view may bypass a lot of complexity.
            # But this needs more study to get right. For now this code will work and
            # we're just going to suppress the warning.  -kmp 16-May-2020
            # Refs:
            #  - https://stackoverflow.com/questions/30102767/custom-route-predicates-in-pyramid
            #  - https://docs.pylonsproject.org/projects/pyramid/en/latest/_modules/pyramid/config/routes.html
            #  - https://docs.pylonsproject.org/projects/pyramid/en/master/narr/hooks.html#view-and-route-predicates
            #  - https://docs.pylonsproject.org/projects/pyramid/en/latest/api/config.html
            custom_predicates=[is_static_page],
            request_method="GET"
        )
    config.add_view(static_page, route_name='staticpage')


def get_pyramid_http_exception_for_redirect_code(code):
    code_dict = {
        301 : HTTPMovedPermanently,
        302 : HTTPFound,
        303 : HTTPSeeOther,
        307 : HTTPTemporaryRedirect
    }
    return code_dict[code]


def generate_page_tree(request, page_name = None):

    current_page_route_root = None
    if page_name is not None:
        page_name_parts = [ path_component for path_component in page_name.split('/') if path_component ]
        if len(page_name_parts) > 1:
            current_page_route_root = [ path_component for path_component in page_name.split('/') if path_component ][0]

    root = { "name" : "", "children" : [], "@id" : '/', "display_title" : "Home" }

    for page in get_iterable_search_results(
        request,
        search_path='/search/',
        param_lists={
            'type'              : ['Page'],
            'sort'              : ['name', 'uuid'],
            'field'             : ['name', 'uuid', '@id', 'display_title', 'children', 'content.name', 'content.title', 'content.title', 'content.@id', 'description', 'redirect.enabled']
        }
    ):
        path_components = [ path_component for path_component in page['name'].split('/') if path_component ]
        if len(path_components) != 1:
            continue
        if current_page_route_root is not None and current_page_route_root == page['name']:
            page.update(request.embed(page['@id'], '@@embedded', as_user=True))
        child_node = {
            "name"          : path_components[0],
            "children"      : page.get('children',[]),
            "@id"           : page['@id'],
            "uuid"          : page['uuid'],
            "display_title" : page['display_title']
        }
        for optional_field in ['description', 'content', 'redirect']:
            if page.get(optional_field) is not None:
                child_node[optional_field] = page[optional_field]
        root['children'].append(child_node)

    cleanup_page_tree(root)

    return root

def cleanup_page_tree(node):
    node['@id'] = '/' + node['name']
    node['@type'] = generate_at_type_for_page(node)
    node_children_length = len(node.get('children', []))
    if node_children_length == 0:
        if node.get('children') is not None:
            del node['children']
        node['is_leaf'] = True
    else:
        node['children'] = [ c for c in node['children'] if c.get('error') is None ]
        for child in node['children']:
            cleanup_page_tree(child)
        node['children'] = [ c for c in node['children'] if ( len(c.get('children', [])) != 0 or len(c.get('content',[])) != 0) ]
        for child_idx, child in enumerate(node['children']):
            child['sibling_length'] = len(node['children'])
            child['sibling_position'] = child_idx

def generate_at_type_for_page(node):
    capitalized_path_names = [ pg.capitalize() for pg in filter(lambda pg: pg, node['@id'].split('/')) ]
    page_type = []
    capitalized_path_names_len = len(capitalized_path_names)
    for cap_idx in range(0, capitalized_path_names_len):
        typestr = ''
        for idx in range(0, capitalized_path_names_len - cap_idx):
            typestr += capitalized_path_names[idx]
        typestr += 'Page'
        page_type.append(typestr)
    if len(node.get('children',[])) > 0:
        page_type.append('DirectoryPage')
    page_type.extend(["StaticPage", "Portal"])
    return page_type

def add_sibling_parent_relations_to_tree(node):
    '''After this is run, will not be able to use with json.dumps() due to (many) new circular relations.'''
    if len(node.get('children', [])) > 0:
        new_children = []
        for idx, child in enumerate(node.get('children', [])):
            new_child = child.copy()
            new_child['parent'] = node
            if idx > 0:
                new_child['previous'] = node['children'][idx - 1]
            if idx < len(node['children']) - 1:
                new_child['next'] = node['children'][idx + 1]
            new_children.append(add_sibling_parent_relations_to_tree(new_child))
        new_node = node.copy()
        new_node['children'] = new_children
        return new_node
    return node


def is_static_page(info, request):
    page_name = "/".join(info.get('match', {}).get('subpath'))
    if '@@' in page_name:
        return False

    path_parts = page_name.split('/')
    if path_parts[0] != "pages" and path_parts[0] in request.registry[COLLECTIONS].keys():
        return False

    request.set_property(lambda x: generate_page_tree(x, page_name), name='_static_page_tree', reify=True)
    request.set_property(lambda x: request.registry[CONNECTION].storage.get_by_json('name', page_name, 'page'), name='_static_page_model', reify=True)

    if request._static_page_model:
        return True
    else:
        return False


def validate_unique_page_name(context, request):
    '''validator to ensure page 'name' lookup_key is unique
    '''
    data = request.json
    # name is required; validate_item_content_post/put/patch will handle missing field
    if 'name' in data:
        lookup_res = request.registry[CONNECTION].storage.get_by_json('name', data['name'], 'page')
        if lookup_res:
            # check_only + POST happens on GUI edit; we cannot confirm if found
            # item is the same item. Let the PATCH take care of validation
            if request.method == 'POST' and request.params.get('check_only', False):
                return
            # editing an item will cause it to find itself. That's okay
            if hasattr(context, 'uuid') and getattr(lookup_res, 'uuid', None) == context.uuid:
                return
            error_msg = ("page %s already exists with name '%s'. This field must be unique"
                         % (lookup_res.uuid, data['name']))
            request.errors.add('body', 'Page: non-unique name',  error_msg)
            return


@view_config(context=Page.Collection, permission='add', request_method='POST',
             validators=[validate_item_content_post, validate_unique_page_name])
@view_config(context=Page.Collection, permission='add_unvalidated', request_method='POST',
             validators=[no_validate_item_content_post],
             request_param=['validate=false'])
@debug_log
def page_add(context, request, render=None):
    return collection_add(context, request, render)


@view_config(context=Page, permission='edit', request_method='PUT',
             validators=[validate_item_content_put, validate_unique_page_name])
@view_config(context=Page, permission='edit', request_method='PATCH',
             validators=[validate_item_content_patch, validate_unique_page_name])
@view_config(context=Page, permission='edit_unvalidated', request_method='PUT',
             validators=[no_validate_item_content_put],
             request_param=['validate=false'])
@view_config(context=Page, permission='edit_unvalidated', request_method='PATCH',
             validators=[no_validate_item_content_patch],
             request_param=['validate=false'])
@view_config(context=Page, permission='index', request_method='GET',
             validators=[validate_item_content_in_place, validate_unique_page_name],
             request_param=['check_only=true'])
@debug_log
def page_edit(context, request, render=None):
    return item_edit(context, request, render)


#### Static Page Routing/Endpoint

def static_page(request):
    '''
    basically get the page in a standard way (item_view_page) which will
    do permissions checking.  Then format the return result to be something
    the front-end expects
    '''

    def remove_relations_in_tree(node, keep="children"):
        '''Returns (deep-)copy'''
        filtered_node = { "name" : node['name'] }
        if keep == 'children' and node.get('children') is not None:
            filtered_node['children'] = [ remove_relations_in_tree(c, keep) for c in node['children'] ]
        if keep == 'parent' and node.get('parent'):
            filtered_node['parent'] = remove_relations_in_tree(node['parent'], keep)
        for field in node.keys():
            if field not in ['next', 'previous', 'children', 'parent'] and node.get(field) is not None:
                filtered_node[field] = node[field]
        return filtered_node

    path_parts = [ path_part for path_part in request.subpath if path_part ]
    page_name = "/".join(path_parts)

    tree = request._static_page_tree

    if tree is not None:

        tree = add_sibling_parent_relations_to_tree(tree)

        curr_node = tree
        page_in_tree = True
        for path_idx, part in enumerate(path_parts):
            for child in curr_node.get('children',[]):
                split_child_name = child['name'].split('/')
                if len(split_child_name) > path_idx and split_child_name[path_idx] == part:
                    curr_node = child
                    break
            if path_idx == len(path_parts) - 1 and curr_node.get('uuid') is None:
                page_in_tree = False
                break
    else:
        page_in_tree = False

    context = Page(request.registry, request._static_page_model)

    if context.properties.get('redirect') and context.properties['redirect'].get('enabled'): # We have a redirect defined.
        parsed_redirect_uri = urlparse(context.properties['redirect'].get('target', '/'))
        uri_to_use = (parsed_redirect_uri.scheme and (parsed_redirect_uri.scheme + ':') or '') + '//' if parsed_redirect_uri.netloc else ''
        uri_to_use += parsed_redirect_uri.path
        uri_to_use += '?' + urlencode({ 'redirected_from' : '/' + context.properties.get('name', str(context.uuid)) }) + ((parsed_redirect_uri.query and ('&' + parsed_redirect_uri.query)) or '')
        # Fallback to 307 as is 'safest' (response isn't cached by browsers)
        return get_pyramid_http_exception_for_redirect_code(context.properties['redirect'].get('code', 307))(location=uri_to_use, detail="Redirected from " + page_name)
    item = item_view_page(context, request)
    cleanup_page_tree(item)
    item['toc'] = item.get('table-of-contents')
    item['@context'] = item['@id']

    if page_in_tree:
        if curr_node.get('next'):
            item['next'] =      remove_relations_in_tree(curr_node['next'])
        if curr_node.get('previous'):
            item['previous'] =  remove_relations_in_tree(curr_node['previous'])
        if curr_node.get('parent'):
            item['parent'] =    remove_relations_in_tree(curr_node['parent'], keep="parent")
        if curr_node.get('sibling_length') is not None:
            item['sibling_length'] = curr_node['sibling_length']
            item['sibling_position'] = curr_node['sibling_position']

    return item


def test_types_file_format(testapp, file_formats):
    """ Tests loading the file_formats fixture, which posts many formats to the app """
    formats = file_formats
    for atid in formats.keys():
        assert testapp.get(f'/file-formats/{atid}', status=200)



from snovault import (
    calculated_property,
    collection,
    load_schema,
)
from snovault.attachment import ItemWithAttachment
from snovault.types.base import (
    Item
)


@collection(
    name='documents',
    properties={
        'title': 'Documents',
        'description': 'Listing of Documents',
    })
class Document(ItemWithAttachment):
    """Document class."""

    item_type = 'document'
    schema = load_schema('encoded_core:schemas/document.json')
    embedded_list = []
    mimetype_map = {'application/proband+xml': ['text/plain']}

    def mimetypes_are_equal(self, m1, m2):
        """ Checks that mime_type m1 and m2 are equal """
        major1 = m1.split('/')[0]
        major2 = m2.split('/')[0]
        if major1 == 'text' and major2 == 'text':
            return True
        if m1 in self.mimetype_map and m2 in self.mimetype_map[m1]:
            return True
        return m1 == m2

    @calculated_property(schema={
        "title": "Display Title",
        "description": "Document filename, if available.",
        "type": "string"
    })
    def display_title(self, attachment=None):
        if attachment:
            return attachment.get('download')
        return Item.display_title(self)

    class Collection(Item.Collection):
        pass



"""The type file for image collection."""

from snovault import (
    collection,
    load_schema,
)
from snovault.types.base import (
    Item,
)
from snovault.attachment import ItemWithAttachment


@collection(
    name='images',
    unique_key='image:filename',
    properties={
        'title': 'Image',
        'description': 'Listing of portal images',
    })
class Image(ItemWithAttachment):
    """Class image,defines accepted file types."""

    item_type = 'image'
    schema = load_schema('encoded_core:schemas/image.json')
    schema['properties']['attachment']['properties']['type']['enum'] = [
        'image/png',
        'image/jpeg',
        'image/gif',
    ]
    embedded_list = Item.embedded_list  # + lab_award_attribution_embed_list

    def unique_keys(self, properties):
        """smth."""
        keys = super(Image, self).unique_keys(properties)
        value = properties.get("attachment", {}).get("download")
        if value:
            keys.setdefault('image:filename', []).append(value)
        return keys



"""The type file for the collection Quality Metric."""
from snovault import (
    abstract_collection,
    collection,
    load_schema,
)
from snovault.types.base import (
    Item
)


"""OVERALL QAULITY SCORE INFO
All QC objects come with a field 'overall_quality_status', which is by default set to 'PASS'
For some qc object we don't have a current protocol to judge the overall quality based on the
fields in the qc item.
When there is a way to make this assesment, add this algorithm as a function to the corresponding
qc class, and update the value. If you implement it for a class with existing items, you will need
to trigger the update with empty patches."""


# NOTE: past version sof quality metrics had many schemas, but now we use 'QualityMetricGeneric' as
# the general way of implementing QCs - should cover 90% of cases, if new types are needed they should
# extend quality metric

@abstract_collection(
    name='quality-metrics',
    properties={
        'title': 'Quality Metrics',
        'description': 'Listing of quality metrics',
    })
class QualityMetric(Item):
    """Quality metrics class."""
    item_type = 'quality_metric'
    base_types = ['QualityMetric'] + Item.base_types
    schema = load_schema('encoded_core:schemas/quality_metric.json')
    embedded_list = Item.embedded_list


@collection(
    name="quality-metrics-generic",
    properties={
        "title": "Generic Quality Metrics",
        "description": "Listing of Generic Quality Metrics",
    },
)
class QualityMetricGeneric(QualityMetric):
    item_type = "quality_metric_generic"
    schema = load_schema("encoded_core:schemas/quality_metric_generic.json")
    embedded_list = QualityMetric.embedded_list



from snovault import collection, load_schema
from snovault.types.base import Item


@collection(
    name='higlass-view-configs',
    unique_key='higlass_view_config:name',
    properties={
        'title': 'HiGlass Displays',
        'description': 'Displays and view configurations for HiGlass',
    })
class HiglassViewConfig(Item):
    """
    Item type which contains a `view_config` property and other metadata.
    """

    item_type = 'higlass_view_config'
    schema = load_schema('encoded_core:schemas/higlass_view_config.json')
    embedded_list = []
    name_key = 'name'



def test_types_software(testapp, software):
    """ Tests that we can load and retrieve a software item """
    software_atid = software.json['@graph'][0]['@id']
    assert testapp.get(software_atid, status=200)



import pytest

from unittest import mock
from ..types import file as tf
from ..types.file import external_creds


@pytest.fixture
def file(testapp, file_formats):
    item = {
        'file_format': file_formats.get('fastq').get('uuid'),
        'md5sum': '00000000000000000000000000000000',
        'filename': 'my.fastq.gz',
        'status': 'uploaded',
    }
    res = testapp.post_json('/file_fastq', item)
    return res.json['@graph'][0]


def test_external_creds():

    with mock.patch.object(tf, 'boto3', autospec=True):

        ret = external_creds('test-wfout-bucket', 'test-key', 'name')
        assert ret['key'] == 'test-key'
        assert ret['bucket'] == 'test-wfout-bucket'
        assert ret['service'] == 's3'
        assert 'upload_credentials' in ret.keys()


@pytest.fixture
def processed_file_data(file_formats):
    return {
        'file_format': file_formats.get('bam').get('uuid'),
    }


def test_validate_produced_from_files_no_produced_by_and_filename_no_filename(
        testapp, processed_file_data):
    res = testapp.post_json('/files-processed', processed_file_data, status=201)
    assert not res.json.get('errors')


def test_validate_filename_invalid_file_format_post(testapp, processed_file_data):
    processed_file_data['file_format'] = 'stringy file format'
    processed_file_data['filename'] = 'test_file.bam'
    res = testapp.post_json('/files-processed', processed_file_data, status=422)
    errors = res.json['errors']
    descriptions = ''.join([e['description'] for e in errors])
    assert 'Problem getting file_format for test_file.bam' in descriptions


def test_validate_filename_valid_file_format_and_name_post(testapp, processed_file_data):
    processed_file_data['filename'] = 'test_file.bam'
    res = testapp.post_json('/files-processed', processed_file_data, status=201)
    assert not res.json.get('errors')


def test_validate_filename_invalid_filename_post(testapp, processed_file_data):
    processed_file_data['filename'] = 'test_file_bam'
    res = testapp.post_json('/files-processed', processed_file_data, status=422)
    errors = res.json['errors']
    descriptions = ''.join([e['description'] for e in errors])
    assert "Filename test_file_bam extension does not agree with specified file format. Valid extension(s): '.bam'" in descriptions


def test_validate_filename_valid_filename_patch(testapp, processed_file_data):
    processed_file_data['filename'] = 'test_file1.bam'
    res1 = testapp.post_json('/files-processed', processed_file_data, status=201)
    assert not res1.json.get('errors')
    res1_props = res1.json['@graph'][0]
    assert res1_props['filename'] == 'test_file1.bam'
    filename2patch = 'test_file2.bam'
    res2 = testapp.patch_json(res1_props['@id'], {'filename': filename2patch}, status=200)
    assert not res2.json.get('errors')
    assert res2.json['@graph'][0]['filename'] == 'test_file2.bam'


def test_validate_filename_invalid_filename_patch(testapp, processed_file_data):
    processed_file_data['filename'] = 'test_file1.bam'
    res1 = testapp.post_json('/files-processed', processed_file_data, status=201)
    assert not res1.json.get('errors')
    res1_props = res1.json['@graph'][0]
    assert res1_props['filename'] == 'test_file1.bam'
    filename2patch = 'test_file2.txt'
    res2 = testapp.patch_json(res1_props['@id'], {'filename': filename2patch}, status=422)
    errors = res2.json['errors']
    descriptions = ''.join([e['description'] for e in errors])
    assert "Filename test_file2.txt extension does not agree with specified file format. Valid extension(s): '.bam'" in descriptions


def test_validate_produced_from_files_invalid_post(testapp, processed_file_data):
    fids = ['not_a_file_id', 'definitely_not']
    processed_file_data['produced_from'] = fids
    res = testapp.post_json('/files-processed', processed_file_data, status=422)
    errors = res.json['errors']
    descriptions = [e['description'] for e in errors]
    for fid in fids:
        desc = "'%s' not found" % fid
        assert desc in descriptions


def test_validate_extra_files_no_extra_files(testapp, processed_file_data):
    res = testapp.post_json('/files-processed', processed_file_data, status=201)
    assert not res.json.get('errors')


def test_validate_extra_files_extra_files_good_post(testapp, processed_file_data):
    extf = {'file_format': 'bai'}
    processed_file_data['extra_files'] = [extf]
    res = testapp.post_json('/files-processed', processed_file_data, status=201)
    assert not res.json.get('errors')


def test_validate_extra_files_extra_files_bad_post_extra_same_as_primary(testapp, processed_file_data):
    extf = {'file_format': 'bam'}
    processed_file_data['extra_files'] = [extf]
    res = testapp.post_json('/files-processed', processed_file_data, status=422)
    assert res.json['errors'][0]['name'] == 'File: invalid extra_file formats'
    assert "'bam' format cannot be the same for file and extra_file" == res.json['errors'][0]['description']


def test_validate_extra_files_extra_files_bad_patch_extra_same_as_primary(testapp, processed_file_data):
    extf = {'file_format': 'bam'}
    res1 = testapp.post_json('/files-processed', processed_file_data, status=201)
    pfid = res1.json['@graph'][0]['@id']
    res2 = testapp.patch_json(pfid, {'extra_files': [extf]}, status=422)
    assert res2.json['errors'][0]['name'] == 'File: invalid extra_file formats'
    assert "'bam' format cannot be the same for file and extra_file" == res2.json['errors'][0]['description']


def test_validate_extra_files_extra_files_bad_post_existing_extra_format(testapp, processed_file_data):
    extfs = [{'file_format': 'bai'}, {'file_format': 'bai'}]
    processed_file_data['extra_files'] = extfs
    res = testapp.post_json('/files-processed', processed_file_data, status=422)
    assert res.json['errors'][0]['name'] == 'File: invalid extra_file formats'
    assert "Multple extra files with 'bai' format cannot be submitted at the same time" == res.json['errors'][0]['description']


def test_validate_extra_files_extra_files_ok_patch_existing_extra_format(testapp, processed_file_data):
    extf = {'file_format': 'bai'}
    processed_file_data['extra_files'] = [extf]
    res1 = testapp.post_json('/files-processed', processed_file_data, status=201)
    pfid = res1.json['@graph'][0]['@id']
    res2 = testapp.patch_json(pfid, {'extra_files': [extf]}, status=200)
    assert not res2.json.get('errors')


def test_validate_extra_files_parent_should_not_have_extras(
        testapp, processed_file_data, file_formats):
    extf = {'file_format': 'bai'}
    processed_file_data['file_format'] = file_formats.get('zip').get('uuid')
    processed_file_data['extra_files'] = [extf]
    res1 = testapp.post_json('/files-processed', processed_file_data, status=422)
    errors = res1.json['errors']
    descriptions = ''.join([e['description'] for e in errors])
    assert "File with format zip should not have extra_files" in descriptions


def test_validate_extra_files_bad_extras_format(
        testapp, processed_file_data, file_formats):
    extf = {'file_format': 'whosit'}
    processed_file_data['extra_files'] = [extf]
    res1 = testapp.post_json('/files-processed', processed_file_data, status=422)
    errors = res1.json['errors']
    descriptions = ''.join([e['description'] for e in errors])
    assert "'whosit' not a valid or known file format" in descriptions


def test_validate_file_format_validity_for_file_type_allows(testapp, file_formats):
    my_fastq_file = {
        'file_format': file_formats.get('fastq').get('uuid'),
    }
    my_proc_file = {
        'file_format': file_formats.get('bam').get('uuid'),
    }
    res1 = testapp.post_json('/files-submitted', my_fastq_file, status=201)
    res2 = testapp.post_json('/files-processed', my_proc_file, status=201)
    assert not res1.json.get('errors')
    assert not res2.json.get('errors')


def test_validate_file_format_validity_for_file_type_fires(testapp, file_formats):
    my_fastq_file = {
        'file_format': file_formats.get('bam').get('uuid'),
    }
    my_proc_file = {
        'file_format': file_formats.get('fastq').get('uuid'),
    }
    res1 = testapp.post_json('/files-submitted', my_fastq_file, status=422)
    errors = res1.json['errors']
    descriptions = ''.join([e['description'] for e in errors])
    assert "File format bam is not allowed for FileSubmitted" in descriptions
    res2 = testapp.post_json('/files-processed', my_proc_file, status=422)
    errors = res2.json['errors']
    descriptions = ''.join([e['description'] for e in errors])
    assert "File format fastq is not allowed for FileProcessed" in descriptions


def test_file_format_does_not_exist(testapp, file_formats):
    my_fastq_file = {
        'file_format': 'waldo',
    }
    res1 = testapp.post_json('/files-submitted', my_fastq_file, status=422)
    errors = res1.json['errors']
    descriptions = ''.join([e['description'] for e in errors])
    assert "'waldo' not found" in descriptions


def test_filename_patch_fails_wrong_format(testapp, file_formats):
    my_fastq_file = {
        'file_format': file_formats.get('fastq').get('uuid'),
        'filename': 'test.fastq.gz'
    }
    res1 = testapp.post_json('/files-submitted', my_fastq_file, status=201)
    resobj = res1.json['@graph'][0]
    patch_data = {"file_format": file_formats.get('bam').get('uuid')}
    res2 = testapp.patch_json('/files-submitted/' + resobj['uuid'], patch_data, status=422)
    errors = res2.json['errors']
    error1 = "Filename test.fastq.gz extension does not agree with specified file format. Valid extension(s): '.bam'"
    error2 = "File format bam is not allowed for FileSubmitted"
    descriptions = ''.join([e['description'] for e in errors])
    assert error1 in descriptions
    assert error2 in descriptions


def test_filename_patch_works_with_different_format(testapp, file_formats):
    my_proc_file = {
        'file_format': file_formats.get('bam').get('uuid'),
        'filename': 'test.bam'
    }
    res1 = testapp.post_json('/files-processed', my_proc_file, status=201)
    resobj = res1.json['@graph'][0]
    patch_data = {"file_format": file_formats.get('zip').get('uuid'), 'filename': 'test.zip'}
    res2 = testapp.patch_json('/files-processed/' + resobj['uuid'], patch_data, status=200)
    assert not res2.json.get('errors')


def test_file_format_patch_works_if_no_filename(testapp, file_formats):
    my_proc_file = {
        'file_format': file_formats.get('bam').get('uuid')
    }
    res1 = testapp.post_json('/files-processed', my_proc_file, status=201)
    resobj = res1.json['@graph'][0]
    patch_data = {"file_format": file_formats.get('zip').get('uuid')}
    res2 = testapp.patch_json('/files-processed/' + resobj['uuid'], patch_data, status=200)
    assert not res2.json.get('errors')



from snovault import (
    calculated_property,
    collection,
    load_schema,
)
from .file import File, _build_file_embedded_list, file_workflow_run_embeds


def _build_file_submitted_embedded_list():
    """Embedded list for FileSubmitted items."""
    return _build_file_embedded_list() + file_workflow_run_embeds + [
        "quality_metric.overall_quality_status",
        "quality_metric.Total Sequences",
        "quality_metric.Sequence length",
        "quality_metric.url",
    ]


@collection(
    name="files-submitted",
    properties={
        "title": "Submitted Files",
        "description": "Listing of Submitted Files",
    })
class FileSubmitted(File):
    """Collection for individual submitted files."""
    item_type = 'file_submitted'
    schema = load_schema('encoded_core:schemas/file_submitted.json')
    embedded_list = _build_file_submitted_embedded_list()
    rev = dict(File.rev, **{
        'workflow_run_inputs': ('WorkflowRun', 'input_files.value'),
        'workflow_run_outputs': ('WorkflowRun', 'output_files.value'),
    })

    @calculated_property(schema={
        "title": "Input of Workflow Runs",
        "description": "All workflow runs that this file serves as an input to",
        "type": "array",
        "items": {
            "title": "Input of Workflow Run",
            "type": ["string", "object"],
            "linkTo": "WorkflowRun"
        }
    })
    def workflow_run_inputs(self, request):
        return self.rev_link_atids(request, "workflow_run_inputs")

    @calculated_property(schema={
        "title": "Output of Workflow Runs",
        "description": "All workflow runs that this file serves as an output from",
        "type": "array",
        "items": {
            "title": "Output of Workflow Run",
            "type": "string",
            "linkTo": "WorkflowRun"
        }
    })
    def workflow_run_outputs(self, request):
        return self.rev_link_atids(request, "workflow_run_outputs")




"""The type file for the collection Software.

moving this out of __init.py__ and into it's own file as
add logic for autopopulating 'name' upon update or create
"""
from snovault import (
    collection,
    load_schema,
)
from snovault.types.base import (
    Item,
)


@collection(
    name='softwares',
    properties={
        'title': 'Softwares',
        'description': 'Listing of software for 4DN analyses',
    })
class Software(Item):
    """The Software class that contains the software... used."""
    item_type = 'software'
    schema = load_schema('encoded_core:schemas/software.json')
    embedded_list = Item.embedded_list  # + lab_award_attribution_embed_list

    def _update(self, properties, sheets=None):
        # update self first to ensure 'software_relation' are stored in self.properties
        if 'version' in properties:
            properties['title'] = properties['name'].replace(' ', '-').lower() + '_' + properties['version']
        elif 'commit' in properties:
            properties['title'] = properties['name'].replace(' ', '-').lower() + '_' + properties['commit']
        else:
            properties['title'] = properties['name'].replace(' ', '-').lower()
        super(Software, self)._update(properties, sheets)

        DicRefRelation = {
            "derived from": "parent of",
            "parent of": "derived from",
            "container for": "contained in",
            "contained in": "container for"
        }
        acc = str(self.uuid)
        if 'software_relation' in properties.keys():
            for relation in properties["software_relation"]:
                switch = relation["relationship_type"]
                rev_switch = DicRefRelation[switch]
                related_sw = relation["software"]
                relationship_entry = {"relationship_type": rev_switch, "software": acc}
                rel_dic = {'software_relation': [relationship_entry, ]}

                target_sw = self.collection.get(related_sw)
                # case one we don't have relations
                if 'software_relation' not in target_sw.properties.keys():
                    target_sw.properties.update(rel_dic)
                    target_sw.update(target_sw.properties)
                else:
                    # case two we have relations but not the one we need
                    for target_relation in target_sw.properties['software_relation']:
                        if target_relation['software'] == acc:
                            break
                    else:
                        # make data for new software_relation
                        target_sw.properties['software_relation'].append(relationship_entry)
                        target_sw.update(target_sw.properties)



def test_types_workflow(testapp, file_formats, workflows):
    """ Tests that we can build workflows that reference file format items """
    for workflow in workflows:
        assert testapp.get(f'/{workflow["@graph"][0]["@id"]}', status=200)


def test_types_workflow_run(testapp, file_formats, sample_file, workflows, sample_workflow_run):
    """ Tests that we can create a workflowrun from an existing workflow """
    assert testapp.get(f'/{sample_workflow_run["@graph"][0]["@id"]}', status=200)



def test_types_metaworkflow(testapp, file_formats, workflows, meta_workflow):
    """ Tests that we can create a metaworkflow using backing structures """
    mwf = meta_workflow
    assert testapp.get(mwf['@graph'][0]['@id'], status=200)



"""
The type file for the collection Pages.  Which is used for static pages on the portal
"""
from snovault import collection, load_schema
from snovault.types.base import Item


@collection(
    name='pages',
    unique_key='page:name',
    properties={
        'title': 'Pages',
        'description': 'Static Pages for the Portal',
    })
class Page(Item):
    """Links to StaticSections"""
    item_type = 'page'
    schema = load_schema('encoded_core:schemas/page.json')
    embedded_list = ['content.*']

    class Collection(Item.Collection):
        pass


# XXX: Legacy code... should not touch it for now... - Will 27 May 2023
for field in ['display_title', 'name', 'description', 'content.name']:
    Page.embedded_list = Page.embedded_list + [
        'children.' + field, 'children.children.' + field, 'children.children.children.' + field
    ]



from snovault import (
    calculated_property,
    collection,
    load_schema,
)
from snovault.attachment import ItemWithAttachment


@collection(
    name='file-formats',
    unique_key='file_format:file_format',
    properties={
        'title': 'File Formats',
        'description': 'Listing of file formats used by 4DN'
    }
)
class FileFormat(ItemWithAttachment):
    """The class to store information about 4DN file formats"""
    item_type = 'file_format'
    schema = load_schema('encoded_core:schemas/file_format.json')
    name_key = 'file_format'

    @calculated_property(schema={
        "title": "Display Title",
        "description": "File Format name or extension.",
        "type": "string"
    })
    def display_title(self, file_format):
        return file_format



"""Abstract collection for UserContent and sub-classes of StaticSection, HiglassViewConfig, etc."""

from snovault import (
    abstract_collection,
    calculated_property,
    collection,
    load_schema
)
from snovault.interfaces import STORAGE
from snovault.types.base import (
    Item,
)
import os
import requests


@abstract_collection(
    name='user-contents',
    unique_key='user_content:name',
    properties={
        'title': "User Content Listing",
        'description': 'Listing of all types of content which may be created by people.',
    })
class UserContent(Item):
    item_type = 'user_content'
    base_types = ['UserContent'] + Item.base_types
    schema = load_schema('encoded_core:schemas/user_content.json')
    embedded_list = []

    @calculated_property(schema={
        "title": "Content",
        "description": "Content (unused)",
        "type": "string"
    })
    def content(self, request):
        return None

    @calculated_property(schema={
        "title": "File Type",
        "description": "Type of this Item (unused)",
        "type": "string"
    })
    def filetype(self, request):
        return None

    def _update(self, properties, sheets=None):
        if properties.get('name') is None and self.uuid is not None:
            properties['name'] = str(self.uuid)
        super(UserContent, self)._update(properties, sheets)

    @classmethod
    def create(cls, registry, uuid, properties, sheets=None):
        submitted_by_uuid = properties.get('submitted_by')
        institution_schema = cls.schema and cls.schema.get('properties', {}).get('institution')
        project_schema = cls.schema and cls.schema.get('properties', {}).get('project')
        if (
            not submitted_by_uuid                               # Shouldn't happen
            or (not institution_schema and not project_schema)            # If not applicable for Item type (shouldn't happen as props defined on UserContent schema)
            or ('institution' in properties or 'project' in properties)   # If values exist already - ideal case - occurs for general submission process(es)
        ):
            # Default for all other Items
            return super(UserContent, cls).create(registry, uuid, properties, sheets)

        submitted_by_item = registry[STORAGE].get_by_uuid(submitted_by_uuid)

        if submitted_by_item:
            # All linkTo property values, if present, are UUIDs
            if 'institution' not in properties and 'institution' in submitted_by_item.properties:
                # Use institution of submitter
                properties['institution'] = submitted_by_item.properties['institution']

            if 'project' not in properties and 'project' in submitted_by_item.properties:
                # Use project of submitter
                properties['project'] = submitted_by_item.properties['project']

        return super(UserContent, cls).create(registry, uuid, properties, sheets)


@collection(
    name='static-sections',
    unique_key='user_content:name',
    properties={
        'title': 'Static Sections',
        'description': 'Static Sections for the Portal',
    })
class StaticSection(UserContent):
    """The Software class that contains the software... used."""
    item_type = 'static_section'
    schema = load_schema('encoded_core:schemas/static_section.json')

    @calculated_property(schema={
        "title": "Content",
        "description": "Content for the page",
        "type": "string"
    })
    def content(self, request, body=None, file=None):

        if isinstance(body, str) or isinstance(body, dict) or isinstance(body, list):
            # Don't need to load in anything. We don't currently support dict/json body (via schema) but could in future.
            return body

        if isinstance(file, str):
            if file[0:4] == 'http' and '://' in file[4:8]:  # Remote File
                return get_remote_file_contents(file)
            else:                                           # Local File
                file_path = os.path.abspath(os.path.dirname(os.path.realpath(__file__)) + "/../../.." + file)   # Go to top of repo, append file
                return get_local_file_contents(file_path)

        return None

    @calculated_property(schema={
        "title": "File Type",
        "description": "Type of file used for content",
        "type": "string"
    })
    def filetype(self, request, body=None, file=None, options=None):
        if options and options.get('filetype') is not None:
            return options['filetype']
        if isinstance(body, str):
            return 'txt'
        if isinstance(body, dict) or isinstance(body, list):
            return 'json'
        if isinstance(file, str):
            filename_parts = file.split('.')
            if len(filename_parts) > 1:
                return filename_parts[len(filename_parts) - 1]
            else:
                return 'txt' # Default if no file extension.
        return None


def get_local_file_contents(filename, contentFilesLocation=None):
    if contentFilesLocation is None:
        full_file_path = filename
    else:
        full_file_path = contentFilesLocation + '/' + filename
    if not os.path.isfile(full_file_path):
        return None
    file = open(full_file_path, encoding="utf-8")
    output = file.read()
    file.close()
    return output


def get_remote_file_contents(uri):
    resp = requests.get(uri)
    return resp.text



from dcicutils.project_utils import C4ProjectRegistry
from snovault.project_defs import SnovaultProject


@C4ProjectRegistry.register('encoded-core')
class EncodedCoreProject(SnovaultProject):
    NAMES = {'NAME': 'core', 'PYPI_NAME': 'encoded-core'}
    ACCESSION_PREFIX = 'ENC'



import pytest
from uuid import uuid4


@pytest.fixture(scope='session')
def file_formats(testapp):
    formats = {}
    ef_format_info = {
        'bai': {'standard_file_extension': 'bam.bai',
                "valid_item_types": ["FileProcessed"]},
        'beddb': {"standard_file_extension": "beddb",
                  "valid_item_types": ["FileProcessed", "FileReference"]},
    }
    format_info = {
        'fastq': {'standard_file_extension': 'fastq.gz',
                  'other_allowed_extensions': ['fq.gz'],
                  "valid_item_types": ["FileFastq", "FileSubmitted"]},
        'bam': {'standard_file_extension': 'bam',
                'extrafile_formats': ['bai'],
                "valid_item_types": ["FileProcessed"]},
        'zip': {'standard_file_extension': 'zip',
                "valid_item_types": ["FileProcessed"]},
        'chromsizes': {'standard_file_extension': 'chrom.sizes',
                       "valid_item_types": ["FileReference"]},
        'other': {'standard_file_extension': '',
                  "valid_item_types": ["FileProcessed", "FileReference"]},
        'bw': {'standard_file_extension': 'bw',
               "valid_item_types": ["FileProcessed"]},
        'bg': {'standard_file_extension': 'bedGraph.gz',
               "valid_item_types": ["FileProcessed"]},
        'bigbed': {'standard_file_extension': 'bb',
                   "valid_item_types": ["FileProcessed", "FileReference"]},
        'bed': {"standard_file_extension": "bed.gz",
                "extrafile_formats": ['beddb'],
                "valid_item_types": ["FileProcessed", "FileReference"]},
        'vcf_gz': {"standard_file_extension": "vcf.gz",
                   "valid_item_types": ["FileProcessed", "FileSubmitted"]}
    }

    for eff, info in ef_format_info.items():
        info['file_format'] = eff
        info['uuid'] = str(uuid4())
        formats[eff] = testapp.post_json('/file_format', info, status=[201]).json['@graph'][0]
    for ff, info in format_info.items():
        info['file_format'] = ff
        info['uuid'] = str(uuid4())
        if info.get('extrafile_formats'):
            eff2add = []
            for eff in info.get('extrafile_formats'):
                eff2add.append(formats[eff].get('@id'))
            info['extrafile_formats'] = eff2add
        formats[ff] = testapp.post_json('/file_format', info, status=[201]).json['@graph'][0]
    return formats


@pytest.fixture(scope='session')
def software(testapp):
    """ Example software item for BWA """
    return testapp.post_json('/software', {
            "status": "shared",
            "software_type": [
                "aligner"
            ],
            "name": "bwa",
            "version": "0.7.17",
            "title": "bwa_0.7.17",
            "uuid": "02d636b9-d82d-4da9-950c-2ca994a13209",
        }, status=201)


@pytest.fixture(scope='session')
def workflows(testapp, file_formats):
    """ Posts many workflow items to the app to be tested """
    workflow_items = [
        {
            "name": "xtea_germline_v1.0.0",
            "title": "xtea germline [v1.0.0]",
            "app_name": "xtea_germline",
            "arguments": [
                {
                    "argument_type": "Input file",
                    "argument_format": "/file-formats/bam",
                    "workflow_argument_name": "input_bam"
                }
            ],
            "app_version": "v1.0.0",
            "description": "Run xTea algorithm to detect Transposable Elements insertion. Implementation to run on germline data.",
            "cwl_main_filename": "xtea_germline.cwl",
            "workflow_language": "CWL",
            "cwl_child_filenames": [],
            "cwl_directory_url_v1": "s3://cgap-msa-main-application-tibanna-cwls/xtea_germline/v1.0.0",
            "uuid": "fcdad1d6-80b7-4acd-bac8-d954540883c4"
        },
        {
            "name": "expansionhunter_germline_v1.0.0",
            "title": "expansionhunter germline [v1.0.0]",
            "app_name": "expansionhunter_germline",
            "accession": "GAPWFJLR56E8",
            "arguments": [
                {
                    "argument_type": "Input file",
                    "argument_format": "/file-formats/bam",
                    "workflow_argument_name": "input_bam"
                },
            ],
            "app_version": "v1.0.0",
            "description": "Run ExpansionHunter algorithm to detect Short Tandem Repeats expansion. Implementation to run on germline data.",
            "cwl_main_filename": "expansionhunter_germline.cwl",
            "workflow_language": "CWL",
            "cwl_child_filenames": [],
            "cwl_directory_url_v1": "s3://cgap-msa-main-application-tibanna-cwls/expansionhunter_germline/v1.0.0",
            "uuid": "fcd43d09-7d50-4c49-acb4-a5accabe9164"
        }
    ]
    res = []
    for workflow in workflow_items:
        res.append(testapp.post_json('/workflow', workflow, status=201).json)
    return res


@pytest.fixture(scope='session')
def meta_workflow(testapp, file_formats, workflows):
    """ Composes a (nonsense) MetaWorkflow from the previously posted WorkFlow items """
    mwf = {
        "name": "xtea_germline_all",
        "input": [{
            "argument_name": "input_bams",
            "argument_type": "file",
            "dimensionality": 1
            },
        ],
        "title": "xtea germline all [v1.0.0]",
        "version": "v1.0.0",
        "accession": "GAPMWT4NX2PA",
        "workflows": [
            {
                "name": "xtea_germline@LINE1",
                "input": [{
                    "scatter": 1,
                    "argument_name": "input_bam",
                    "argument_type": "file",
                    "source_argument_name": "input_bams"
                    },
                ],
                "config": {
                    "ebs_size": "1.1x",
                    "run_name": "run_xtea_germline-LINE1",
                    "EBS_optimized": True,
                    "instance_type": "m5.4xlarge",
                    "spot_instance": True,
                    "behavior_on_capacity_limit": "wait_and_retry"
                },
                "workflow": "fcdad1d6-80b7-4acd-bac8-d954540883c4",
                "custom_pf_fields": {}
            },
        ],
        "description": "Run xTea algorithm to detect Transposable Elements insertion from germline data. "
                       "Call ALL classes of TE (LINE1, ALU, SVA).",
        "uuid": "cc3fbaef-4dd3-482b-afc9-01bfcd3fd2b4"
    }
    return testapp.post_json('/MetaWorkflow', mwf, status=201).json


@pytest.fixture(scope='session')
def sample_file(testapp, file_formats):
    """ Posts a sample file we can reference in workflow output """
    return testapp.post_json('/FileProcessed', {
        'uuid': 'fe58c408-cc82-4ece-8062-b2a4a81c745d',
        'file_format': '/file-formats/bam'
    }, status=201).json


@pytest.fixture(scope='session')
def sample_workflow_run(testapp, file_formats, workflows, sample_file):
    """ Posts a sample workflowrun """
    return testapp.post_json('/WorkflowRunAwsem', {
        "title": "xtea v1.1.0 run 2023-04-17 21:02:32.148147",
        "workflow": "fcdad1d6-80b7-4acd-bac8-d954540883c4",
        "parameters": [],
        "run_status": "complete",
        "input_files": [
            {
                "value": "fe58c408-cc82-4ece-8062-b2a4a81c745d",
                "ordinal": 1,
                "dimension": "0",
                "workflow_argument_name": "input_bam"
            }
        ],
        "awsem_job_id": "bOhzpCaK16oj",
        "run_platform": "AWSEM",
        "metadata_only": False,
        "awsem_app_name": "xtea",
        "uuid": "4e15dd9d-2f3b-4444-8e7b-f7ec3044669e"
    }, status=201).json



import datetime
import pytz
from urllib.parse import parse_qs, urlparse
from pyramid.settings import asbool
from pyramid.view import view_config
from pyramid.httpexceptions import (
    HTTPTemporaryRedirect,
    HTTPNotFound,
)
from snovault.util import build_s3_presigned_get_url
from .types.quality_metric import QualityMetric


def includeme(config):
    config.scan(__name__)


# S3 URL identifier
S3_BUCKET_DOMAIN_SUFFIX = '.s3.amazonaws.com'


def parse_qc_s3_url(url):
    """ Parses the given s3 URL into its pair of bucket, key
        Note that this function works the way it does because of how these
        urls end up in our database. Eventually we should clean this up.
        TODO: use version in utils
        Format:
            https://s3.amazonaws.com/cgap-devtest-main-application-cgap-devtest-wfout/GAPFI1HVXJ5F/fastqc_report.html
            https://cgap-devtest-main-application-tibanna-logs.s3.amazonaws.com/41c2fJDQcLk3.metrics/metrics.html
    """
    parsed_url = urlparse(url)
    if parsed_url.hostname.endswith(S3_BUCKET_DOMAIN_SUFFIX):
        bucket = parsed_url.hostname[:-len(S3_BUCKET_DOMAIN_SUFFIX)]
        key = parsed_url.path.lstrip('/')
    else:
        [bucket, key] = parsed_url.path.lstrip('/').split('/', 1)
    return bucket, key


@view_config(name='download', context=QualityMetric, request_method='GET',
             permission='view', subpath_segments=[0, 1])
def download(context, request):
    """ Downloads the quality metric report from S3 """
    properties = context.upgrade_properties()
    if 'url' not in properties:
        raise HTTPNotFound(properties)
    # parse direct s3 link
    # format: https://s3.amazonaws.com/cgap-devtest-main-application-cgap-devtest-wfout/GAPFI1HVXJ5F/fastqc_report.html
    # or: https://cgap-devtest-main-application-tibanna-logs.s3.amazonaws.com/41c2fJDQcLk3.metrics/metrics.html
    bucket, key = parse_qc_s3_url(properties['url'])
    params_to_get_obj = {
        'Bucket': bucket,
        'Key': key
    }
    location = build_s3_presigned_get_url(params=params_to_get_obj)

    if asbool(request.params.get('soft')):
        expires = int(parse_qs(urlparse(location).query)['Expires'][0])
        return {
            '@type': ['SoftRedirect'],
            'location': location,
            'expires': datetime.datetime.fromtimestamp(expires, pytz.utc).isoformat(),
        }

    # 307 redirect specifies to keep original method
    raise HTTPTemporaryRedirect(location=location)  # 307


import boto3
import json
import logging
import os
import structlog
import transaction

from botocore.exceptions import ClientError
from copy import deepcopy
from dcicutils.ecr_utils import CGAP_ECR_REGION
from pyramid.threadlocal import get_current_request
from pyramid.traversal import resource_path
from dcicutils.secrets_utils import assume_identity
from dcicutils.misc_utils import override_environ
from snovault import (
    CONNECTION,
    calculated_property,
    load_schema,
    abstract_collection,
)
from snovault.invalidation import add_to_indexing_queue
from snovault.util import make_s3_client
from snovault.types.base import (
    Item,
    get_item_or_none,
)


logging.getLogger('boto3').setLevel(logging.CRITICAL)


log = structlog.getLogger(__name__)

file_workflow_run_embeds = [
    'workflow_run_inputs.workflow.title',
    'workflow_run_inputs.input_files.workflow_argument_name',
    'workflow_run_inputs.input_files.value.filename',
    'workflow_run_inputs.input_files.value.display_title',
    'workflow_run_inputs.input_files.value.file_format',
    'workflow_run_inputs.input_files.value.uuid',
    'workflow_run_inputs.input_files.value.accession',
    'workflow_run_inputs.output_files.workflow_argument_name',
    'workflow_run_inputs.output_files.value.display_title',
    'workflow_run_inputs.output_files.value.file_format',
    'workflow_run_inputs.output_files.value.uuid',
    'workflow_run_inputs.output_files.value.accession',
    'workflow_run_inputs.output_files.value_qc.url',
    'workflow_run_inputs.output_files.value_qc.overall_quality_status'
]

file_workflow_run_embeds_processed = (file_workflow_run_embeds
                                      + [e.replace('workflow_run_inputs.', 'workflow_run_outputs.')
                                         for e in file_workflow_run_embeds])


def show_upload_credentials(request=None, context=None, status=None):
    if request is None or status not in ('uploading', 'to be uploaded by workflow', 'upload failed'):
        return False
    return request.has_permission('edit', context)


def external_creds(bucket, key, name=None, profile_name=None):
    """
    if name is None, we want the link to s3 but no need to generate
    an access token.  This is useful for linking metadata to files that
    already exist on s3.
    """

    logging.getLogger('boto3').setLevel(logging.CRITICAL)
    credentials = {}
    s3_encrypt_key_id = None  # might be reassigned later from identity.get('ENCODED_S3_ENCRYPT_KEY_ID')
    if name is not None:
        policy = {
            'Version': '2012-10-17',
            'Statement': [
                {
                    'Effect': 'Allow',
                    'Action': 's3:PutObject',
                    'Resource': f'arn:aws:s3:::{bucket}/{key}',
                },
            ]
        }
        # In the new environment, extract S3 Keys from global application configuration
        if 'IDENTITY' in os.environ:
            identity = assume_identity()
            with override_environ(**identity):
                conn = boto3.client('sts',
                                    aws_access_key_id=os.environ.get('S3_AWS_ACCESS_KEY_ID'),
                                    aws_secret_access_key=os.environ.get('S3_AWS_SECRET_ACCESS_KEY'))
            s3_encrypt_key_id = identity.get('ENCODED_S3_ENCRYPT_KEY_ID')
            if s3_encrypt_key_id:  # must be used with ACCOUNT_NUMBER as well
                policy['Statement'].append({  # NoQA - PyCharm doesn't like this append for some bogus reason
                    'Effect': 'Allow',
                    'Action': [
                        'kms:Encrypt',
                        'kms:Decrypt',
                        'kms:ReEncrypt*',
                        'kms:GenerateDataKey*',
                        'kms:DescribeKey'
                    ],
                    'Resource': f'arn:aws:kms:{CGAP_ECR_REGION}:{identity["ACCOUNT_NUMBER"]}:key/{s3_encrypt_key_id}'
                })
        # In the old account, we are always passing IAM User creds so these will just work
        else:
            conn = boto3.client('sts',
                                aws_access_key_id=os.environ.get('AWS_ACCESS_KEY_ID'),
                                aws_secret_access_key=os.environ.get('AWS_SECRET_ACCESS_KEY'))
        token = conn.get_federation_token(Name=name, Policy=json.dumps(policy))
        # 'access_key' 'secret_key' 'expiration' 'session_token'
        credentials = token.get('Credentials')
        # Convert Expiration datetime object to string via cast
        # Uncaught serialization error picked up by Docker - Will 2/25/2021
        credentials['Expiration'] = str(credentials['Expiration'])
        credentials.update({
            'upload_url': f's3://{bucket}/{key}',
            'federated_user_arn': token.get('FederatedUser').get('Arn'),
            'federated_user_id': token.get('FederatedUser').get('FederatedUserId'),
            's3_encrypt_key_id': s3_encrypt_key_id,
            'request_id': token.get('ResponseMetadata').get('RequestId'),
            'key': key
        })
    return {
        'service': 's3',
        'bucket': bucket,
        'key': key,
        'upload_credentials': credentials,
    }


def property_closure(request, propname, root_uuid):
    # Must avoid cycles
    conn = request.registry[CONNECTION]
    seen = set()
    remaining = {str(root_uuid)}
    while remaining:
        seen.update(remaining)
        next_remaining = set()
        for uuid in remaining:
            obj = conn.get_by_uuid(uuid)
            next_remaining.update(obj.__json__(request).get(propname, ()))
        remaining = next_remaining - seen
    return seen


def _build_file_embedded_list():
    """Embedded list for File type."""
    return [
        # FileFormat linkTo
        "file_format.file_format",

        # File linkTo
        "related_files.relationship_type",
        "related_files.file.accession",
        "related_files.file.file_format.file_format",

        # QC
        "quality_metric.@type",
        "quality_metric.qc_list.qc_type",
        "quality_metric.qc_list.value.uuid",
    ]


@abstract_collection(
    name='files',
    unique_key='accession',
    properties={
        'title': 'Files',
        'description': 'Listing of Files',
    })
class File(Item):
    """Collection for individual files."""
    item_type = 'file'
    base_types = ['File'] + Item.base_types
    schema = load_schema('encoded_core:schemas/file.json')
    embedded_list = _build_file_embedded_list()

    @calculated_property(schema={
        "title": "Display Title",
        "description": "Name of this File",
        "type": "string"
    })
    def display_title(self, request, file_format, accession=None, external_accession=None):
        accession = accession or external_accession
        file_format_item = get_item_or_none(request, file_format, 'file-formats')
        try:
            file_extension = '.' + file_format_item.get('standard_file_extension')
        except AttributeError:
            file_extension = ''
        return '{}{}'.format(accession, file_extension)

    @calculated_property(schema={
        "title": "File Type",
        "description": "Type of File",
        "type": "string"
    })
    def file_type_detailed(self, request, file_format, file_type=None):
        outString = (file_type or 'other')
        file_format_item = get_item_or_none(request, file_format, 'file-formats')
        try:
            fformat = file_format_item.get('file_format')
            outString = outString + ' (' + fformat + ')'
        except AttributeError:
            pass
        return outString

    # def generate_track_title(self, track_info, props):
    #     if not props.get('higlass_uid'):
    #         return None
    #     exp_type = track_info.get('experiment_type', None)
    #     if exp_type is None:
    #         return None
    #     bname = track_info.get('biosource_name', 'unknown sample')
    #     ftype = props.get('file_type', 'unspecified type')
    #     assay = track_info.get('assay_info', '')
    #
    #     title = '{ft} for {bs} {et} {ai}'.format(
    #         ft=ftype, ai=assay, et=exp_type, bs=bname
    #     )
    #     return title.replace('  ', ' ').rstrip()

    def _get_file_expt_bucket(self, request, item2check):
        fatid = self.jsonld_id(request)
        if 'files' in item2check:
            if fatid in item2check.get('files'):
                return 'raw file'
        if 'processed_files' in item2check:
            if fatid in item2check.get('processed_files'):
                return 'processed file'
        of_info = item2check.get('other_processed_files', [])
        for obucket in of_info:
            ofiles = obucket.get('files')
            if [of for of in ofiles if of == fatid]:
                return obucket.get('title')
        return None

    def _update(self, properties, sheets=None):
        if not properties:
            return
        # ensure we always have s3 links setup
        sheets = {} if sheets is None else sheets.copy()
        uuid = self.uuid
        old_creds = self.propsheets.get('external', None)
        new_creds = old_creds

        # don't get new creds
        if properties.get('status', None) in ('uploading', 'to be uploaded by workflow',
                                              'upload failed'):
            new_creds = self.build_external_creds(self.registry, uuid, properties)
            sheets['external'] = new_creds

        # handle extra files
        updated_extra_files = []
        extra_files = properties.get('extra_files', [])
        if extra_files:
            # get @id for parent file
            try:
                at_id = resource_path(self)
            except Exception:
                at_id = "/" + str(uuid) + "/"
            # ensure at_id ends with a slash
            if not at_id.endswith('/'):
                at_id += '/'

            file_formats = []
            for xfile in extra_files:
                # ensure a file_format (identifier for extra_file) is given and non-null
                if not ('file_format' in xfile and bool(xfile['file_format'])):
                    continue
                eformat = xfile['file_format']
                if eformat.startswith('/file-formats/'):
                    eformat = eformat[len('/file-formats/'):-1]
                xfile_format = self.registry['collections']['FileFormat'].get(eformat)
                xff_uuid = str(xfile_format.uuid)
                if not xff_uuid:
                    raise Exception("Cannot find format item for the extra file")

                if xff_uuid in file_formats:
                    raise Exception("Each file in extra_files must have unique file_format")
                file_formats.append(xff_uuid)
                xfile['file_format'] = xff_uuid

                xfile['accession'] = properties.get('accession')
                # just need a filename to trigger creation of credentials
                xfile_name = xfile.get("filename")
                if xfile_name is None:
                    xfile['filename'] = xfile['accession']
                xfile['uuid'] = str(uuid)
                # if not 'status' in xfile or not bool(xfile['status']):
                #    xfile['status'] = properties.get('status')
                ext = self.build_external_creds(self.registry, uuid, xfile)
                # build href
                file_extension = xfile_format.properties.get('standard_file_extension')
                filename = '{}.{}'.format(xfile['accession'], file_extension)
                xfile['href'] = at_id + '@@download/' + filename
                xfile['upload_key'] = ext['key']
                sheets['external' + xfile['file_format']] = ext
                updated_extra_files.append(xfile)

        if extra_files:
            properties['extra_files'] = updated_extra_files

        if old_creds:
            if old_creds.get('key') != new_creds.get('key'):
                try:
                    conn = make_s3_client()
                    bname = old_creds['bucket']
                    conn.delete_object(Bucket=bname, Key=old_creds['key'])
                except Exception as e:
                    print(e)

        # update self first to ensure 'related_files' are stored in self.properties
        super(File, self)._update(properties, sheets)

        # handle related_files. This is quite janky; must manually invalidate
        # the relation made on `related_fl` item because we are calling its
        # update() method directly, which circumvents snovault.crud_view.item_edit
        DicRefRelation = {
            "derived from": "parent of",
            "parent of": "derived from",
            "supercedes": "is superceded by",
            "is superceded by": "supercedes",
            "paired with": "paired with"
        }

        if 'related_files' in properties:
            my_uuid = str(self.uuid)
            # save these values
            curr_txn = None
            curr_request = None
            for relation in properties["related_files"]:
                try:
                    switch = relation["relationship_type"]
                    rev_switch = DicRefRelation[switch]
                    related_fl = relation["file"]
                    relationship_entry = {"relationship_type": rev_switch, "file": my_uuid}
                except Exception:
                    log.error('Error updating related_files on %s _update. %s'
                              % (my_uuid, relation))
                    continue

                target_fl = self.collection.get(related_fl)
                target_fl_props = deepcopy(target_fl.properties)
                # This is a cool python feature. If break is not hit in the loop,
                # go to the `else` statement. Works for empty lists as well
                for target_relation in target_fl_props.get('related_files', []):
                    if (target_relation.get('file') == my_uuid
                            and target_relation.get('relationship_type') == rev_switch):
                        break
                else:
                    # Get the current request in order to queue the forced
                    # update for indexing. This is bad form.
                    # Don't do this anywhere else, please!
                    if curr_txn is None:
                        curr_txn = transaction.get()
                    if curr_request is None:
                        curr_request = get_current_request()
                    # handle related_files whether or not any currently exist
                    target_related_files = target_fl_props.get('related_files', [])
                    target_related_files.append(relationship_entry)
                    target_fl_props.update({'related_files': target_related_files})
                    target_fl._update(target_fl_props)
                    to_queue = {'uuid': str(target_fl.uuid), 'sid': target_fl.sid,
                                'info': 'queued from %s _update' % my_uuid}
                    curr_txn.addAfterCommitHook(add_to_indexing_queue,
                                                args=(curr_request, to_queue, 'edit'))

    @property
    def __name__(self):
        properties = self.upgrade_properties()
        if properties.get('status') == 'replaced':
            return self.uuid
        return properties.get(self.name_key, None) or self.uuid

    def unique_keys(self, properties):
        keys = super(File, self).unique_keys(properties)
        if properties.get('status') != 'replaced':
            if 'md5sum' in properties:
                value = 'md5:{md5sum}'.format(**properties)
                keys.setdefault('alias', []).append(value)
        return keys

    @calculated_property(schema={
        "title": "Title",
        "type": "string",
        "description": "Accession of this file"
    })
    def title(self, accession=None, external_accession=None):
        return accession or external_accession

    @calculated_property(schema={
        "title": "Download URL",
        "type": "string",
        "description": "Use this link to download this file."
    })
    def href(self, request, file_format, accession=None, external_accession=None):
        fformat = get_item_or_none(request, file_format, 'file-formats')
        try:
            file_extension = '.' + fformat.get('standard_file_extension')
        except AttributeError:
            file_extension = ''
        accession = accession or external_accession
        filename = '{}{}'.format(accession, file_extension)
        return request.resource_path(self) + '@@download/' + filename

    @calculated_property(schema={
        "title": "Upload Key",
        "type": "string",
    })
    def upload_key(self, request):
        properties = self.properties
        external = self.propsheets.get('external', {})
        if not external:
            try:
                external = self.build_external_creds(self.registry, self.uuid, properties)
            except ClientError:
                log.error(os.environ)
                log.error(properties)
                return 'UPLOAD KEY FAILED'
        return external['key']

    @calculated_property(condition=show_upload_credentials, schema={
        "type": "object",
    })
    def upload_credentials(self):
        external = self.propsheets.get('external', None)
        if external is not None:
            return external['upload_credentials']

    @calculated_property(condition=show_upload_credentials, schema={
        "type": "object",
    })
    def extra_files_creds(self):
        external = self.propsheets.get('external', None)
        if external is not None:
            extras = []
            for extra in self.properties.get('extra_files', []):
                eformat = extra.get('file_format')
                xfile_format = self.registry['collections']['FileFormat'].get(eformat)
                try:
                    xff_uuid = str(xfile_format.uuid)
                except AttributeError:
                    print("Can't find required format uuid for %s" % eformat)
                    continue
                extra_creds = self.propsheets.get('external' + xff_uuid)
                extra['upload_credentials'] = extra_creds['upload_credentials']
                extras.append(extra)
            return extras

    @classmethod
    def get_bucket(cls, registry):
        return registry.settings['file_upload_bucket']

    @classmethod
    def build_external_creds(cls, registry, uuid, properties):
        bucket = cls.get_bucket(registry)
        fformat = properties.get('file_format')
        if fformat.startswith('/file-formats/'):
            fformat = fformat[len('/file-formats/'):-1]
        prop_format = registry['collections']['FileFormat'].get(fformat)
        try:
            file_extension = prop_format.properties['standard_file_extension']
        except KeyError:
            raise Exception('File format not in list of supported file types')
        key = '{uuid}/{accession}.{file_extension}'.format(
            file_extension=file_extension, uuid=uuid,
            accession=properties.get('accession'))

        # remove the path from the file name and only take first 32 chars
        fname = properties.get('filename')
        name = None
        if fname:
            name = fname.split('/')[-1][:32]

        profile_name = registry.settings.get('file_upload_profile_name')
        return external_creds(bucket, key, name, profile_name)

    @classmethod
    def create(cls, registry, uuid, properties, sheets=None):
        if properties.get('status') in ('uploading', 'to be uploaded by workflow'):
            sheets = {} if sheets is None else sheets.copy()
            sheets['external'] = cls.build_external_creds(registry, uuid, properties)
        return super(File, cls).create(registry, uuid, properties, sheets)

    class Collection(Item.Collection):
        pass



"""The type file for the collection MetaWorkflow and MetaWorkflowRun.

moving this out of __init.py__ and into it's own file as
add logic for autopopulating 'name' upon update or create
"""
from snovault import (
    collection,
    load_schema,
)
from snovault.types.base import (
    Item,
)


def _build_meta_workflow_run_embedded_list():
    """Create embedded list for MetaWorkflowRuns."""
    return [
        # MetaWorkflow linkTo
        "meta_workflow.title",
        "meta_workflow.version",
        "meta_workflow.workflows.input.argument_name",
        "meta_workflow.workflows.input.source_argument_name",
        "meta_workflow.workflows.input.argument_type",
        "meta_workflow.workflows.input.value",
        "meta_workflow.workflows.input.value_type",
        "meta_workflow.workflows.name",
        "meta_workflow.workflows.workflow.display_title",
        "meta_workflow.workflows.workflow.status",
        "meta_workflow.workflows.workflow.uuid",
        # "workflow_runs.workflow_run.input_files.value.workflow_run_inputs.@id",

        # When part of `input_files`, `value` is a File linkTo/embed
        "workflow_runs.workflow_run.run_status",
        "workflow_runs.workflow_run.input_files.value.workflow_run_outputs.@id",
        "workflow_runs.workflow_run.input_files.value.quality_metric.overall_quality_status",
        "workflow_runs.workflow_run.input_files.value.file_size",
        "workflow_runs.workflow_run.input_files.value.file_format",
        "workflow_runs.workflow_run.input_files.workflow_argument_name",
        "workflow_runs.workflow_run.input_files.*",

        # When part of `parameters`, `value` is a string (maybe number)
        # Removed for now, until we maybe update ReactWorkflowViz to re-use param nodes. Not high priority.
        # "workflow_runs.workflow_run.parameters.value",
        # "workflow_runs.workflow_run.parameters.workflow_argument_name",
        # "workflow_runs.workflow_run.parameters.*",

        "workflow_runs.workflow_run.output_files.value.workflow_run_inputs.@id",
        "workflow_runs.workflow_run.output_files.value.quality_metric.overall_quality_status",
        "workflow_runs.workflow_run.output_files.value.file_size",
        "workflow_runs.workflow_run.output_files.value.file_format",
        "workflow_runs.workflow_run.output_files.workflow_argument_name",
        "workflow_runs.workflow_run.output_files.*",
        # "workflow_runs.workflow_run.output_files.value.workflow_run_outputs.@id"
    ]


@collection(
    name='meta-workflows',
    properties={
        'title': 'MetaWorkflows',
        'description': 'Listing of MetaWorkflows',
    })
class MetaWorkflow(Item):
    item_type = 'meta_workflow'
    schema = load_schema('encoded_core:schemas/meta_workflow.json')


@collection(
    name='meta-workflow-runs',
    properties={
        'title': 'MetaWorkflowRuns',
        'description': 'Listing of MetaWorkflowRuns',
    })
class MetaWorkflowRun(Item):
    item_type = 'meta_workflow_run'
    embedded_list = _build_meta_workflow_run_embedded_list()
    schema = load_schema('encoded_core:schemas/meta_workflow_run.json')



import transaction
from snovault import (
    calculated_property,
    collection,
    load_schema,
    COLLECTIONS
)
from snovault.crud_views import collection_add as sno_collection_add
from snovault.schema_utils import validate_request
from snovault.validation import ValidationFailure
from snovault.types.base import (
    Item,
    DELETED_ACL,
    ONLY_ADMIN_VIEW_ACL
)


@collection(
    name='tracking-items',
    properties={
        'title': 'TrackingItem',
        'description': 'For internal tracking of ENCODED events',
    })
class TrackingItem(Item):
    """tracking-item class."""

    item_type = 'tracking_item'
    schema = load_schema('encoded_core:schemas/tracking_item.json')
    embedded_list = []
    STATUS_ACL = {
        'released': ONLY_ADMIN_VIEW_ACL,
        'deleted': DELETED_ACL,
        'draft': ONLY_ADMIN_VIEW_ACL
    }

    @classmethod
    def create_and_commit(cls, request, properties, clean_headers=False):
        """
        Create a TrackingItem with a given request and properties, committing
        it directly to the DB. This works by manually committing the
        transaction, which may cause issues if this function is called as
        part of another POST. For this reason, this function should be used to
        track GET requests -- otherwise, use the standard POST method.
        If validator issues are hit, will not create the item but log to error

        Args:
            request: current request object
            properties (dict): TrackingItem properties to post
            clean_headers(bool): If True, remove 'Location' header created by POST

        Returns:
            dict response from snovault.crud_views.collection_add

        Raises:
            ValidationFailure if TrackingItem cannot be validated
        """
        collection = request.registry[COLLECTIONS]['TrackingItem']
        # set remote_user to standarize permissions
        prior_remote = request.remote_user
        request.remote_user = 'EMBED'
        # remove any missing attributes from DownloadTracking
        properties['download_tracking'] = {
            k: v for k, v in properties.get('download_tracking', {}).items() if v is not None
        }
        validate_request(collection.type_info.schema, request, properties)
        if request.errors:  # added from validate_request
            request.remote_user = prior_remote
            raise ValidationFailure('body', 'TrackingItem: create_and_commit',
                                    'Cannot validate request')
        ti_res = sno_collection_add(collection, request, False)  # render=False
        transaction.get().commit()
        if clean_headers and 'Location' in request.response.headers:
            del request.response.headers['Location']
        request.remote_user = prior_remote
        return ti_res

    @calculated_property(schema={
        "title": "Display Title",
        "description": "Descriptor of TrackingItem",
        "type": "string"
    })
    def display_title(self, tracking_type, date_created=None, google_analytics=None):
        if date_created:  # pragma: no cover should always be true
            date_created = date_created[:10]
        if tracking_type == 'google_analytics':
            for_date = None
            if google_analytics:
                for_date = google_analytics.get('for_date', None)
            if for_date:
                return 'Google Analytics for ' + for_date
            return 'Google Analytics Item'
        elif tracking_type == 'download_tracking':
            title = 'Download Tracking Item'
            if date_created:
                title = title + ' from ' + date_created
            return title
        else:
            title = 'Tracking Item'
            if date_created:
                title = title + ' from ' + date_created
            return title



_app_settings = {
    "collection_datastore": "database",
    "item_datastore": "database",
    "multiauth.policies": "session remoteuser accesskey auth0",
    "multiauth.groupfinder": "snovault.authorization.groupfinder",
    "multiauth.policy.session.use": "snovault.authentication.NamespacedAuthenticationPolicy",
    "multiauth.policy.session.base": "pyramid.authentication.SessionAuthenticationPolicy",
    "multiauth.policy.session.namespace": "mailto",
    "multiauth.policy.remoteuser.use": "snovault.authentication.NamespacedAuthenticationPolicy",
    "multiauth.policy.remoteuser.namespace": "remoteuser",
    "multiauth.policy.remoteuser.base": "pyramid.authentication.RemoteUserAuthenticationPolicy",
    "multiauth.policy.accesskey.use": "snovault.authentication.NamespacedAuthenticationPolicy",
    "multiauth.policy.accesskey.namespace": "accesskey",
    "multiauth.policy.accesskey.base": "snovault.authentication.BasicAuthAuthenticationPolicy",
    "multiauth.policy.accesskey.check": "snovault.authentication.basic_auth_check",
    "multiauth.policy.auth0.use": "snovault.authentication.NamespacedAuthenticationPolicy",
    "multiauth.policy.auth0.namespace": "auth0",
    "multiauth.policy.auth0.base": "snovault.authentication.Auth0AuthenticationPolicy",
    "load_test_only": True,
    "testing": True,
    "indexer": True,
    "mpindexer": False,
    "production": True,
    "pyramid.debug_authorization": True,
    "postgresql.statement_timeout": 20,
    "sqlalchemy.url": "dummy@dummy",
    "retry.attempts": 3,
    # some file specific stuff for testing
    "file_upload_bucket": "cgap-unit-testing-files",
    "file_wfout_bucket": "cgap-unit-testing-wfout",
    "file_upload_profile_name": "test-profile",
    "metadata_bundles_bucket": "cgap-unit-testing-metadata-bundles",
}


def make_app_settings_dictionary():
    return _app_settings.copy()



import copy
from snovault import calculated_property, collection, load_schema
from snovault.types.base import Item


def _build_workflows_embedded_list():
    """ Helper function for building workflow embedded list. """
    return Item.embedded_list + [
            'steps.name',

            # Objects
            'steps.inputs.*',
            'steps.outputs.*',

            # Software linkTo
            'steps.meta.software_used.name',
            'steps.meta.software_used.title',
            'steps.meta.software_used.version',
            'steps.meta.software_used.source_url',

            # FileFormat linkTo
            'arguments.argument_format.file_format',

            'arguments.argument_type',
            'arguments.workflow_argument_name'
    ]


@collection(
    name='workflows',
    properties={
        'title': 'Workflows',
        'description': 'Listing of 4DN analysis workflows',
    })
class Workflow(Item):
    """The Workflow class that describes a workflow and steps in it."""

    item_type = 'workflow'
    schema = load_schema('encoded_core:schemas/workflow.json')
    embedded_list = _build_workflows_embedded_list()
    rev = {
        'newer_versions': ('Workflow', 'previous_version')
    }

    @calculated_property(schema={
        "title": "Newer Versions",
        "description": "Newer versions of this workflow",
        "type": "array",
        "exclude_from": ["FFedit-create"],
        "items": {
            "title": "Newer versions",
            "type": ["string", "object"],
            "linkTo": "Workflow"
        }
    })
    def newer_versions(self, request):
        return self.rev_link_atids(request, "newer_versions")


def _build_workflow_run_embedded_list():
    """ Helper function for building workflow embedded list. """
    return Item.embedded_list + [
        # Workflow linkTo
        'workflow.category',
        'workflow.experiment_types',
        'workflow.app_name',
        'workflow.title',
        'workflow.steps.name',

        # Software linkTo
        'workflow.steps.meta.software_used.name',
        'workflow.steps.meta.software_used.title',
        'workflow.steps.meta.software_used.version',
        'workflow.steps.meta.software_used.source_url',

        # String
        'input_files.workflow_argument_name',
        # File linkTo
        'input_files.value.filename',
        'input_files.value.display_title',
        'input_files.value.href',
        'input_files.value.file_format',
        'input_files.value.accession',
        'input_files.value.@type',
        'input_files.value.@id',
        'input_files.value.file_size',
        'input_files.value.quality_metric.url',
        'input_files.value.quality_metric.overall_quality_status',
        'input_files.value.status',

        # String
        'output_files.workflow_argument_name',

        # File linkTo
        'output_files.value.filename',
        'output_files.value.display_title',
        'output_files.value.href',
        'output_files.value.file_format',
        'output_files.value.accession',
        'output_files.value.@type',
        'output_files.value.@id',
        'output_files.value.file_size',
        'output_files.value.quality_metric.url',
        'output_files.value.quality_metric.overall_quality_status',
        'output_files.value.status',
        'output_files.value_qc.url',
        'output_files.value_qc.overall_quality_status'
    ]


steps_run_data_schema = {
    "type": "object",
    "properties": {
        "file": {
            "type": "array",
            "title": "File(s)",
            "description": "File(s) for this step input/output argument.",
            "items": {
                "type": ["string", "object"],  # Either string (uuid) or a object/dict containing uuid & other front-end-relevant properties from File Item.
                "linkTo": "File"  # TODO: (Med/High Priority) Make this work. Will likely wait until after embedding edits b.c. want to take break from WF stuff and current solution works.
            }
        },
        "meta": {
            "type": "array",
            "title": "Additional metadata for input/output file(s)",
            "description": "List of additional info that might be related to file, but not part of File Item itself, such as ordinal.",
            "items": {
                "type": "object"
            }
        },
        "value": {  # This is used in place of run_data.file, e.g. for a parameter string value, that does not actually have a file.
            "title": "Value",
            "type": "string",
            "description": "Value used for this output argument."
        },
        "type": {
            "type": "string",
            "title": "I/O Type"
        }
    }
}


workflow_schema = load_schema('encoded_core:schemas/workflow.json')
workflow_steps_property_schema = workflow_schema.get('properties', {}).get('steps')
workflow_run_steps_property_schema = copy.deepcopy(workflow_steps_property_schema)
workflow_run_steps_property_schema['items']['properties']['inputs']['items']['properties']['run_data'] = steps_run_data_schema
workflow_run_steps_property_schema['items']['properties']['outputs']['items']['properties']['run_data'] = steps_run_data_schema


@collection(
    name='workflow-runs',
    properties={
        'title': 'Workflow Runs',
        'description': 'Listing of executions of 4DN analysis workflows',
    })
class WorkflowRun(Item):
    """The WorkflowRun class that describes execution of a workflow."""

    item_type = 'workflow_run'
    schema = load_schema('encoded_core:schemas/workflow_run.json')
    embedded_list = _build_workflow_run_embedded_list()

    @calculated_property(schema=workflow_run_steps_property_schema, category='page')
    def steps(self, request):
        '''
        Extends the 'inputs' & 'outputs' (lists of dicts) properties of calculated property 'analysis_steps' (list of dicts) from
        WorkflowRun's related Workflow with additional property 'run_data', which contains references to Files, Parameters, and Reports
        generated by this specific Workflow Run.

        :returns: List of analysis_steps items, extended with 'inputs' and 'outputs'.
        '''
        workflow = self.properties.get('workflow')
        if workflow is None:
            return []

        workflow = request.embed('/' + workflow)
        analysis_steps = workflow.get('steps')

        if not isinstance(analysis_steps, list) or len(analysis_steps) == 0:
            return []

        # fileCache = {} # Unnecessary unless we'll convert file @id into plain embedded dictionary, in which case we use this to avoid re-requests for same file UUID.

        def get_global_source_or_target(all_io_source_targets):
            # Find source or target w/o a 'step'.
            # Step outputs or inputs with a source or target without a "step" defined
            # are considered global inputs/outputs. Matching WorkflowRun.[output|input]_files
            # is done against step step.[inputs | output].[target | source].name.
            global_pointing_source_target = [
                source_target for source_target in all_io_source_targets
                if source_target.get('step') is None
            ]
            if len(global_pointing_source_target) > 1:
                raise Exception('Found more than one source or target without a step.')
            if len(global_pointing_source_target) == 0:
                return None
            return global_pointing_source_target[0]


        def map_run_data_to_io_arg(step_io_arg, wfr_runtime_inputs, io_type):
            '''
            Add file metadata in form of 'run_data' : { 'file' : { '@id', 'display_title', etc. } } to AnalysisStep dict's 'input' or 'output' list item dict
            if one of own input or output files' workflow_argument_name matches the AnalysisStep dict's input or output's sourceOrTarget.workflow_argument_name.

            :param step_io_arg: Reference to an 'input' or 'output' dict passed in from a Workflow-derived analysis_step.
            :param wfr_runtime_inputs: List of Step inputs or outputs, such as 'input_files', 'output_files', 'quality_metric', or 'parameters'.
            :returns: True if found and added run_data property to analysis_step.input or analysis_step.output (param inputOrOutput).
            '''
            # is_global_arg = step_io_arg.get('meta', {}).get('global', False) == True
            # if not is_global_arg:
            #    return False # Skip. We only care about global arguments.

            value_field_name = 'value' if io_type == 'parameter' else 'file'

            global_pointing_source_target = get_global_source_or_target(step_io_arg.get('source', step_io_arg.get('target', [])))
            if not global_pointing_source_target:
                return False

            matched_runtime_io_data = [
                io_object for io_object in wfr_runtime_inputs
                if (
                    global_pointing_source_target['name'] == io_object.get('workflow_argument_name') and
                    # Quality Metrics might be saved with `value_qc` in place of `value`
                    (io_object.get('value') is not None or io_object.get('value_qc') is not None)
                )
            ]

            if len(matched_runtime_io_data) > 0:
                matched_runtime_io_data = sorted(matched_runtime_io_data, key=lambda io_object: io_object.get('ordinal', 1))

                # List of file or Item UUIDs.
                value_list = []

                # Aligned-to-file-list-indices list of file metadata(s)
                meta_list = []

                for io_dict in matched_runtime_io_data:
                    # Quality Metrics might be saved with `value_qc` instead of `value`
                    linked_to_item_uuid = io_dict.get('value') or io_dict.get('value_qc')
                    value_list.append(linked_to_item_uuid)
                    # Add all remaining properties from dict in (e.g.) 'input_files','output_files',etc. list.
                    # Contains things like dimension, ordinal, file_format, and so forth.
                    meta_dict = { k:v for (k,v) in io_dict.items() if k not in [ 'value', 'value_qc', 'type', 'workflow_argument_name' ] }
                    # There is a chance we do not have the file_format in the input_files list. Most often this would occur if multiple
                    # files in input argument list
                    meta_list.append(meta_dict)

                step_io_arg['run_data'] = {
                    value_field_name : value_list,
                    "type" : io_type,
                    "meta" : meta_list
                }
                return True
            return False


        def mergeArgumentsWithSameArgumentName(args):
            '''Merge arguments with the same workflow_argument_name, unless differing ordinals'''
            seen_argument_names = {}
            resultArgs = []
            for arg in args:
                argName = arg.get('workflow_argument_name')
                if argName:
                    priorArgument = seen_argument_names.get(argName)
                    if priorArgument and priorArgument.get('ordinal', 1) == arg.get('ordinal', 1):
                        priorArgument.update(arg)
                    else:
                        resultArgs.append(arg)
                        seen_argument_names[argName] = arg
            return resultArgs


        output_files = mergeArgumentsWithSameArgumentName(self.properties.get('output_files', []))
        input_files = mergeArgumentsWithSameArgumentName(self.properties.get('input_files', []))
        input_params = mergeArgumentsWithSameArgumentName(self.properties.get('parameters', []))

        for step in analysis_steps:
            # Add output file metadata to step outputs & inputs, based on workflow_argument_name v step output target name.

            for output in step['outputs']:
                map_run_data_to_io_arg(output, output_files, 'output')

            for input in step['inputs']:
                if input.get('meta', {}).get('type') != 'parameter':
                    map_run_data_to_io_arg(input, input_files, 'input')
                else:
                    map_run_data_to_io_arg(input, input_params, 'parameter')

        return analysis_steps


@collection(
    name='workflow-runs-awsem',
    properties={
        'title': 'Workflow Runs AWSEM',
        'description': 'Listing of executions of 4DN analysis workflows on AWSEM platform',
    })
class WorkflowRunAwsem(WorkflowRun):
    """The WorkflowRun class that describes execution of a workflow on AWSEM platform."""
    base_types = ['WorkflowRun'] + Item.base_types
    item_type = 'workflow_run_awsem'
    schema = load_schema('encoded_core:schemas/workflow_run_awsem.json')
    embedded_list = WorkflowRun.embedded_list



from snovault import (
    collection,
    load_schema,
)
from .file import File


@collection(
    name='files-reference',
    properties={
        'title': 'Reference Files',
        'description': 'Listing of Reference Files',
        })
class FileReference(File):
    """Collection for individual reference files."""
    item_type = 'file_reference'
    schema = load_schema('encoded_core:schemas/file_reference.json')
    embedded_list = File.embedded_list



"""py.test fixtures for Pyramid.

http://pyramid.readthedocs.org/en/latest/narr/testing.html
"""

import logging
import pkg_resources
import pytest
import webtest
from sqlalchemy import exc
# from dcicutils.ff_mocks import NO_SERVER_FIXTURES

from dcicutils.qa_utils import notice_pytest_fixtures, MockFileSystem
from pyramid.request import apply_request_extensions
from pyramid.testing import DummyRequest
from pyramid.threadlocal import get_current_registry, manager as threadlocal_manager
from snovault import DBSESSION, ROOT, UPGRADER
from snovault.elasticsearch import ELASTIC_SEARCH, create_mapping
from snovault.util import generate_indexer_namespace_for_testing
from snovault.loadxl import load_all
from .conftest_settings import make_app_settings_dictionary
from .. import main


"""
README:
    * This file contains application level fixtures and hooks in the server/data fixtures present in
      other files. 
    * There are "app" based fixtures that rely only on postgres, "es_app" fixtures that 
      use both postgres and ES (for search/ES related testing)
"""


# This should work but does not seem to... various issues related to rollbacks occurring or not
# occurring when they should/shouldn't, probably related to zsa_savepoints
# @pytest.fixture
# def external_tx(request, conn):
#     # overridden from snovault to detect and continue from savepoint error
#     if NO_SERVER_FIXTURES:
#         yield 'NO_SERVER_FIXTURES'
#         return
#
#     notice_pytest_fixtures(request)
#     with conn.begin_nested() as tx:
#         yield tx


# hacked version
@pytest.fixture
def external_tx(request, conn):
    # overridden from snovault to detect and continue from savepoint error
    # if NO_SERVER_FIXTURES:
    #     yield 'NO_SERVER_FIXTURES'
    #     return

    notice_pytest_fixtures(request)
    # print('BEGIN external_tx')
    try:
        tx = conn.begin_nested()
    except exc.PendingRollbackError as e:
        if 'inactive savepoint transaction' in str(e):
            conn._nested_transaction.rollback()
            tx = conn.begin_nested()
        else:
            raise
    try:
        yield tx
        tx.rollback()
    except exc.InternalError as e:
        if 'savepoint' in str(e) and 'does not exist' in str(e):
            pass
        else:
            raise

    # conn does not implement .rollback()

    # The database should be empty unless a data fixture was loaded
    # for table in Base.metadata.sorted_tables:
    #     assert conn.execute(table.count()).scalar() == 0


@pytest.fixture(autouse=True)
def autouse_external_tx(external_tx):
    pass


@pytest.fixture(scope='session')
def app_settings(request, wsgi_server_host_port, conn, DBSession):  # noQA - We didn't choose the fixture name.
    notice_pytest_fixtures(request, wsgi_server_host_port, conn, DBSession)
    settings = make_app_settings_dictionary()
    settings['auth0.audiences'] = 'http://%s:%s' % wsgi_server_host_port
    settings[DBSESSION] = DBSession
    return settings


INDEXER_NAMESPACE_FOR_TESTING = generate_indexer_namespace_for_testing('cgap')


@pytest.fixture(scope='session')
def es_app_settings(wsgi_server_host_port, elasticsearch_server, postgresql_server, aws_auth):
    settings = make_app_settings_dictionary()
    settings['create_tables'] = True
    settings['persona.audiences'] = 'http://%s:%s' % wsgi_server_host_port  # 2-tuple such as: ('localhost', '5000')
    settings['elasticsearch.server'] = elasticsearch_server
    settings['sqlalchemy.url'] = postgresql_server
    settings['collection_datastore'] = 'elasticsearch'
    settings['item_datastore'] = 'elasticsearch'
    settings['indexer'] = True
    settings['indexer.namespace'] = INDEXER_NAMESPACE_FOR_TESTING

    # use aws auth to access elasticsearch
    if aws_auth:
        settings['elasticsearch.aws_auth'] = aws_auth
    return settings


def pytest_configure():
    logging.basicConfig(format='%(message)s')
    logging.getLogger('sqlalchemy.engine').setLevel(logging.WARNING)

    class Shorten(logging.Filter):
        max_len = 500

        def filter(self, record):
            if record.msg == '%r':
                record.msg = record.msg % record.args
                record.args = ()
            if len(record.msg) > self.max_len:
                record.msg = record.msg[:self.max_len] + '...'
            return True

    logging.getLogger('sqlalchemy.engine.base.Engine').addFilter(Shorten())


@pytest.fixture
def threadlocals(request, dummy_request, registry):
    notice_pytest_fixtures(request, dummy_request, registry)
    threadlocal_manager.push({'request': dummy_request, 'registry': registry})
    yield dummy_request
    threadlocal_manager.pop()


class MyDummyRequest(DummyRequest):
    def remove_conditional_headers(self):
        pass

    def _get_registry(self):
        if self._registry is None:
            return get_current_registry()
        return self._registry

    def _set_registry(self, registry):
        self.__dict__['registry'] = registry

    def _del_registry(self):
        self._registry = None

    registry = property(_get_registry, _set_registry, _del_registry)


@pytest.fixture
def dummy_request(root, registry, app):
    request = app.request_factory.blank('/dummy')
    request.root = root
    request.registry = registry
    request._stats = {}
    request.invoke_subrequest = app.invoke_subrequest
    apply_request_extensions(request)
    return request


@pytest.fixture(scope='session')
def app(app_settings):
    """ WSGI application level functional testing. """
    return main({}, **app_settings)


@pytest.fixture(scope='session')
def es_app(es_app_settings, **kwargs):
    """
    App that uses both Postgres and ES - pass this as "app" argument to TestApp.
    Pass all kwargs onto create_mapping
    """
    app = main({}, **es_app_settings)
    create_mapping.run(app, **kwargs)

    return app


@pytest.fixture
def registry(app):
    return app.registry


@pytest.fixture
def elasticsearch(registry):
    return registry[ELASTIC_SEARCH]


@pytest.fixture
def upgrader(registry):
    return registry[UPGRADER]


@pytest.fixture
def root(registry):
    return registry[ROOT]


# Available Fixtures
# ------------------
#
#  ################## +-----------------------------------------+----------------------------------------------------+
#  ################## |               Basic Application         |      Application with ES + Postgres                |
#  ################## +-----------------------+-----------------+---------------------------+------------------------+
#  ################## |   JSON content        |  HTML content   |      JSON content         |      HTML content      |
#  -------------------+-----------------------+-----------------+---------------------------+------------------------+
#  Anonymous User     | anontestapp           | anonhtmltestapp |  anon_es_testapp          | anon_html_es_testapp   |
#  -------------------+-----------------------+-----------------+---------------------------+------------------------+
#  System User        | testapp               | htmltestapp     |  es_testapp               | html_es_testapp        |
#  -------------------+-----------------------+-----------------+---------------------------+------------------------+
#  Authenticated User | authenticated_testapp | -----           |  authenticated_es_testapp | -----                  |
#  -------------------+-----------------------+-----------------+---------------------------+------------------------+
#  Submitter User     | submitter_testapp     | -----           |  -----                    | -----                  |
#  -------------------+-----------------------+-----------------+---------------------------+------------------------+
#  Indexer User       | -----                 | -----           |  indexer_testapp          | -----                  |
#  -------------------+-----------------------+-----------------+---------------------------+------------------------+
#  Embed User         | embed_testapp         | -----           |  -----                    | -----                  |
#  -------------------+-----------------------+-----------------+---------------------------+------------------------+
#
# TODO: Reconsider naming to have some underscores interspersed for better readability.
#       e.g., html_testapp rather than htmltestapp, and especially anon_html_test_app rather than anonhtmltestapp.
#       -kmp 03-Feb-2020


@pytest.fixture
def anontestapp(app):
    """TestApp for anonymous user (i.e., no user specified), accepting JSON data."""
    environ = {
        'HTTP_ACCEPT': "application/json"
    }
    return webtest.TestApp(app, environ)


@pytest.fixture
def anonhtmltestapp(app):
    """TestApp for anonymous (not logged in) user, accepting text/html content."""
    environ = {
        'HTTP_ACCEPT': 'text/html'
    }
    test_app = webtest.TestApp(app, environ)
    return test_app


@pytest.fixture
def anon_es_testapp(es_app):
    """ TestApp simulating a bare Request entering the application (with ES enabled) """
    environ = {
        'HTTP_ACCEPT': 'application/json'
    }
    return webtest.TestApp(es_app, environ)


@pytest.fixture
def anon_html_es_testapp(es_app):
    """TestApp with ES + Postgres for anonymous (not logged in) user, accepting text/html content."""
    environ = {
        'HTTP_ACCEPT': 'text/html'
    }
    return webtest.TestApp(es_app, environ)


@pytest.fixture(scope="session")
def testapp(app):
    """TestApp for username TEST, accepting JSON data."""
    # TODO: Use make_testapp from .helpers
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST'
    }
    return webtest.TestApp(app, environ)


@pytest.fixture
def htmltestapp(app):
    """TestApp for TEST user, accepting text/html content."""
    # TODO: Use make_htmltestapp from .helpers
    environ = {
        'HTTP_ACCEPT': 'text/html',
        'REMOTE_USER': 'TEST',
    }
    test_app = webtest.TestApp(app, environ)
    return test_app


@pytest.fixture(scope='session')
def es_testapp(es_app):
    """ TestApp with ES + Postgres. Must be imported where it is needed. """
    # TODO: Use make_testapp from .helpers
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST',
    }
    return webtest.TestApp(es_app, environ)


@pytest.fixture
def html_es_testapp(es_app):
    """TestApp with ES + Postgres for TEST user, accepting text/html content."""
    # TODO: Use make_htmltestapp from .helpers
    environ = {
        'HTTP_ACCEPT': 'text/html',
        'REMOTE_USER': 'TEST',
    }
    return webtest.TestApp(es_app, environ)


@pytest.fixture
def authenticated_testapp(app):
    """TestApp for an authenticated, non-admin user (TEST_AUTHENTICATED), accepting JSON data."""
    # TODO: Use make_authenticated_testapp from .helpers
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST_AUTHENTICATED',
    }
    return webtest.TestApp(app, environ)


@pytest.fixture
def authenticated_es_testapp(es_app):
    """ TestApp for authenticated non-admin user with ES """
    # TODO: Use make_authenticated_testapp from .helpers
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST_AUTHENTICATED',
    }
    return webtest.TestApp(es_app, environ)


@pytest.fixture
def submitter_testapp(app):
    """TestApp for a non-admin user (TEST_SUBMITTER), accepting JSON data."""
    # TODO: Use make_submitter_testapp from .helpers
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST_SUBMITTER',
    }
    return webtest.TestApp(app, environ)


@pytest.fixture
def indexer_testapp(es_app):
    """ Indexer testapp, meant for manually triggering indexing runs by posting to /index.
        Always uses the ES app (obviously, but not so obvious previously) """
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'INDEXER',
    }
    return webtest.TestApp(es_app, environ)


@pytest.fixture
def embed_testapp(app):
    """TestApp for user EMBED, accepting JSON data."""
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'EMBED',
    }
    return webtest.TestApp(app, environ)


@pytest.fixture
def wsgi_app(wsgi_server):
    """TestApp for WSGI server."""
    return webtest.TestApp(wsgi_server)


class WorkbookCache:
    """ Caches whether or not we have already provisioned the workbook. """
    done = None

    @classmethod
    def initialize_if_needed(cls, es_app):
        if not cls.done:
            cls.done = cls.make_fresh_workbook(es_app)

    @classmethod
    def make_fresh_workbook(cls, es_app):
        environ = {
            'HTTP_ACCEPT': 'application/json',
            'REMOTE_USER': 'TEST',
        }
        testapp = webtest.TestApp(es_app, environ)

        # Just load the workbook inserts
        # Note that load_all returns None for success or an Exception on failure.
        load_res = load_all(testapp, pkg_resources.resource_filename('encoded', 'tests/data/workbook-inserts/'), [])

        if isinstance(load_res, Exception):
            raise load_res
        elif load_res:
            raise RuntimeError("load_all returned a true value that was not an exception.")

        # There is a potential timing error here that doing
        #    testapp.post_json('/index', {'record': True})
        # instead would seem to 'fix', even though the record part is not strictly needed.
        # Perhaps because it waits until done, or waits long enough, that indexing completes. -kmp 31-Jan-2023
        testapp.post_json('/index', {})
        return True


@pytest.fixture(scope='session')
def workbook(es_app):
    """ Loads a bunch of data (tests/data/workbook-inserts) into the system on first run
        (session scope doesn't work). """
    WorkbookCache.initialize_if_needed(es_app)


@pytest.fixture
def mocked_file_system():
    with MockFileSystem(auto_mirror_files_for_read=True).mock_exists_open_remove():
        yield
